import functools

import numpy as np
import scipy.sparse.linalg as spla
import scipy.linalg.interpolative as sli
from autoray import (
    do,
    reshape,
    dag,
    infer_backend,
    astype,
    get_dtype_name,
    compose,
    backend_like,
)

from ..core import njit, generated_jit
from ..linalg import base_linalg
from ..linalg import rand_linalg


# some convenience functions for multiplying diagonals


def rdmul(x, d):
    """Right-multiplication a matrix by a vector representing a diagonal."""
    return x * reshape(d, (1, -1))


def rddiv(x, d):
    """Right-multiplication of a matrix by a vector representing an inverse
    diagonal.
    """
    return x / reshape(d, (1, -1))


def ldmul(d, x):
    """Left-multiplication a matrix by a vector representing a diagonal."""
    return x * reshape(d, (-1, 1))


def lddiv(d, x):
    """Left-multiplication of a matrix by a vector representing an inverse
    diagonal.
    """
    return x / reshape(d, (-1, 1))


@njit  # pragma: no cover
def dag_numba(x):
    return np.conjugate(x.T)


@njit  # pragma: no cover
def rdmul_numba(x, d):
    return x * d.reshape(1, -1)


@njit  # pragma: no cover
def rddiv_numba(x, d):
    return x / d.reshape(1, -1)


@njit  # pragma: no cover
def ldmul_numba(d, x):
    return x * d.reshape(-1, 1)


@njit  # pragma: no cover
def lddiv_numba(d, x):
    return x / d.reshape(-1, 1)


@compose
def sgn(x):
    """Get the 'sign' of ``x``, such that ``x / sgn(x)`` is real and
    non-negative.
    """
    return x / (do('abs', x) + (x == 0.0))


@sgn.register("numpy")
@generated_jit
def sgn_numba(x):
    from numba import types
    if hasattr(x, "dtype"):
        dtype = x.dtype
    else:
        dtype = x
    if isinstance(dtype, types.Complex):
        return lambda x: x / (np.abs(x) + (x == 0.0))
    else:
        return lambda x: np.sign(x) + (x == 0.0)


def _trim_and_renorm_svd_result(
    U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
):
    """Give full SVD decomposion result ``U``, ``s``, ``VH``, optionally trim,
    renormalize, and absorb the singular values. See ``svd_truncated`` for
    details.
    """
    if (cutoff > 0.0) or (renorm > 0):
        if cutoff_mode == 1:  # 'abs'
            n_chi = do("count_nonzero", s > cutoff)

        elif cutoff_mode == 2:  # 'rel'
            n_chi = do("count_nonzero", s > cutoff * s[0])

        elif cutoff_mode in (3, 4, 5, 6):
            if cutoff_mode in (3, 4):
                pow = 2
            else:
                pow = 1

            sp = s**pow
            csp = do("cumsum", sp, 0)
            tot = csp[-1]

            if cutoff_mode in (4, 6):
                n_chi = do("count_nonzero", csp < (1 - cutoff) * tot) + 1
            else:
                n_chi = do("count_nonzero", (tot - csp) > cutoff) + 1

        n_chi = max(n_chi, 1)
        if max_bond > 0:
            n_chi = min(n_chi, max_bond)

    elif max_bond > 0:
        # only maximum bond specified
        n_chi = max_bond
    else:
        # neither maximum bond dimension nor cutoff specified
        n_chi = s.shape[0]

    if n_chi < s.shape[0]:
        s = s[:n_chi]
        U = U[:, :n_chi]
        VH = VH[:n_chi, :]

        if renorm > 0:
            norm = (tot / csp[n_chi - 1]) ** (1 / pow)
            s *= norm

    # XXX: tensorflow can't multiply mixed dtypes
    if infer_backend(s) == "tensorflow":
        dtype = get_dtype_name(U)
        if "complex" in dtype:
            s = astype(s, dtype)

    if absorb is None:
        return U, s, VH
    if absorb == -1:
        U = rdmul(U, s)
    elif absorb == 1:
        VH = ldmul(s, VH)
    else:
        s = do("sqrt", s)
        U = rdmul(U, s)
        VH = ldmul(s, VH)

    return U, None, VH


@compose
def svd_truncated(
    x,
    cutoff=-1.0,
    cutoff_mode=3,
    max_bond=-1,
    absorb=0,
    renorm=0,
    backend=None,
):
    """Truncated svd or raw array ``x``.

    Parameters
    ----------
    cutoff : float
        Singular value cutoff threshold, if ``cutoff <= 0.0``, then only
        ``max_bond`` is used.
    cutoff_mode : {1, 2, 3, 4, 5, 6}
        How to perform the trim:

            - 1: ['abs'], trim values below ``cutoff``
            - 2: ['rel'], trim values below ``s[0] * cutoff``
            - 3: ['sum2'], trim s.t. ``sum(s_trim**2) < cutoff``.
            - 4: ['rsum2'], trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
            - 5: ['sum1'], trim s.t. ``sum(s_trim**1) < cutoff``.
            - 6: ['rsum1'], trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

    max_bond : int
        An explicit maximum bond dimension, use -1 for none.
    absorb : {-1, 0, 1, None}
        How to absorb the singular values. -1: left, 0: both, 1: right and
        None: don't absorb (return).
    renorm : {0, 1}
        Whether to renormalize the singular values (depends on `cutoff_mode`).
    """
    with backend_like(backend):
        U, s, VH = do("linalg.svd", x)
        return _trim_and_renorm_svd_result(
            U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
        )


@njit(["i4(f4[:], f4, i4)", "i4(f8[:], f8, i4)"])  # pragma: no cover
def _compute_number_svals_to_keep_numba(s, cutoff, cutoff_mode):
    """Find the number of singular values to keep of ``s`` given ``cutoff`` and
    ``cutoff_mode``.
    """
    if cutoff_mode == 1:  # 'abs'
        n_chi = np.sum(s > cutoff)

    elif cutoff_mode == 2:  # 'rel'
        n_chi = np.sum(s > cutoff * s[0])

    elif cutoff_mode in (3, 4, 5, 6):
        if cutoff_mode in (3, 4):
            pow = 2
        else:
            pow = 1

        target = cutoff
        if cutoff_mode in (4, 6):
            target *= np.sum(s**pow)

        n_chi = s.size
        ssum = 0.0
        for i in range(s.size - 1, -1, -1):
            s2 = s[i] ** pow
            if not np.isnan(s2):
                ssum += s2
            if ssum > target:
                break
            n_chi -= 1

    return max(n_chi, 1)


@njit(["f4(f4[:], i4, f4)", "f8(f8[:], i4, f8)"])  # pragma: no cover
def _compute_svals_renorm_factor_numba(s, n_chi, renorm):
    """Find the normalization constant for ``s`` such that the new sum squared
    of the ``n_chi`` largest values equals the sum squared of all the old ones.
    """
    s_tot_keep = 0.0
    s_tot_lose = 0.0
    for i in range(s.size):
        s2 = s[i] ** renorm
        if not np.isnan(s2):
            if i < n_chi:
                s_tot_keep += s2
            else:
                s_tot_lose += s2
    return ((s_tot_keep + s_tot_lose) / s_tot_keep) ** (1 / renorm)


@njit  # pragma: no cover
def _trim_and_renorm_svd_result_numba(
    U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
):
    """Accelerate version of ``_trim_and_renorm_svd_result``."""
    if (cutoff > 0.0) or (renorm > 0):
        n_chi = _compute_number_svals_to_keep_numba(s, cutoff, cutoff_mode)

        if max_bond > 0:
            n_chi = min(n_chi, max_bond)

        if n_chi < s.size:
            if renorm > 0:
                s = s[:n_chi] * _compute_svals_renorm_factor_numba(
                    s, n_chi, renorm
                )
            else:
                s = s[:n_chi]

            U = U[:, :n_chi]
            VH = VH[:n_chi, :]

    elif (max_bond != -1) and (max_bond < s.shape[0]):
        U = U[:, :max_bond]
        s = s[:max_bond]
        VH = VH[:max_bond, :]

    s = np.ascontiguousarray(s)

    if absorb is None:
        return U, s, VH
    elif absorb == -1:
        U = rdmul_numba(U, s)
    elif absorb == 1:
        VH = ldmul_numba(s, VH)
    else:
        s **= 0.5
        U = rdmul_numba(U, s)
        VH = ldmul_numba(s, VH)

    return U, None, VH


@svd_truncated.register("numpy")
@njit  # pragma: no cover
def svd_truncated_numba(
    x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0, renorm=0
):
    """Accelerated version of ``svd_truncated`` for numpy arrays."""
    U, s, VH = np.linalg.svd(x, full_matrices=False)
    return _trim_and_renorm_svd_result_numba(
        U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


def svdvals(x):
    """SVD-decomposition, but return singular values only."""
    return np.linalg.svd(x, full_matrices=False, compute_uv=False)


@njit  # pragma: no cover
def _svd_via_eig_truncated_numba(
    x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0, renorm=0
):
    """SVD-split via eigen-decomposition."""
    if x.shape[0] > x.shape[1]:
        # Get sU, V
        s2, V = np.linalg.eigh(dag_numba(x) @ x)
        U = x @ V
        VH = dag_numba(V)
        # small negative eigenvalues turn into nan when sqrtd
        s2[s2 < 0.0] = 0.0
        s = np.sqrt(s2)
        U = rddiv_numba(U, s)
    else:
        # Get U, sV
        s2, U = np.linalg.eigh(x @ dag_numba(x))
        VH = dag_numba(U) @ x
        s2[s2 < 0.0] = 0.0
        s = np.sqrt(s2)
        VH = lddiv_numba(s, VH)

    # we need singular values and vectors in descending order
    U, s, VH = U[:, ::-1], s[::-1], VH[::-1, :]

    return _trim_and_renorm_svd_result_numba(
        U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


def svd_via_eig_truncated(
    x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0, renorm=0
):
    if isinstance(x, np.ndarray):
        return _svd_via_eig_truncated_numba(
            x, cutoff, cutoff_mode, max_bond, absorb, renorm
        )

    if x.shape[0] > x.shape[1]:
        # Get sU, V
        s2, V = do("linalg.eigh", dag(x) @ x)
        U = x @ V
        VH = dag(V)
        # small negative eigenvalues turn into nan when sqrtd
        s2 = do("clip", s2, 0.0, None)
        s = do("sqrt", s2)
        U = rddiv(U, s)
    else:
        # Get U, sV
        s2, U = do("linalg.eigh", x @ dag(x))
        VH = dag(U) @ x
        s2 = do("clip", s2, 0.0, None)
        s = do("sqrt", s2)
        VH = lddiv(s, VH)

    # we need singular values and vectors in descending order
    U, s, VH = do("flip", U, (1,)), do("flip", s, (0,)), do("flip", VH, (0,))

    return _trim_and_renorm_svd_result(
        U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


@njit  # pragma: no cover
def svdvals_eig(x):  # pragma: no cover
    """SVD-decomposition via eigen, but return singular values only."""
    if x.shape[0] > x.shape[1]:
        s2 = np.linalg.eigvalsh(dag_numba(x) @ x)
    else:
        s2 = np.linalg.eigvalsh(x @ dag_numba(x))

    s2[s2 < 0.0] = 0.0
    return s2[::-1] ** 0.5


@njit  # pragma: no cover
def eigh(x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0, renorm=0):
    """SVD-decomposition, using hermitian eigen-decomposition, only works if
    ``x`` is hermitian.
    """
    s, U = np.linalg.eigh(x)
    s, U = s[::-1], U[:, ::-1]  # make sure largest singular value first

    V = ldmul_numba(sgn_numba(s), dag_numba(U))
    s = np.abs(s)
    return _trim_and_renorm_svd_result_numba(
        U, s, V, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


def _choose_k(x, cutoff, max_bond):
    """Choose the number of singular values to target."""
    d = min(x.shape)

    if cutoff != 0.0:
        k = rand_linalg.estimate_rank(
            x, cutoff, k_max=None if max_bond < 0 else max_bond
        )
    else:
        k = min(d, max_bond)

    # if computing more than half of spectrum then just use dense method
    return "full" if k > d // 2 else k


def svds(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0, renorm=0):
    """SVD-decomposition using iterative methods. Allows the
    computation of only a certain number of singular values, e.g. max_bond,
    from the get-go, and is thus more efficient. Can also supply
    ``scipy.sparse.linalg.LinearOperator``.
    """
    k = _choose_k(x, cutoff, max_bond)

    if k == "full":
        if not isinstance(x, np.ndarray):
            x = x.to_dense()
        return svd_truncated(x, cutoff, cutoff_mode, max_bond, absorb)

    U, s, VH = base_linalg.svds(x, k=k)
    return _trim_and_renorm_svd_result_numba(
        U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


def isvd(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0, renorm=0):
    """SVD-decomposition using interpolative matrix random methods. Allows the
    computation of only a certain number of singular values, e.g. max_bond,
    from the get-go, and is thus more efficient. Can also supply
    ``scipy.sparse.linalg.LinearOperator``.
    """
    k = _choose_k(x, cutoff, max_bond)

    if k == "full":
        if not isinstance(x, np.ndarray):
            x = x.to_dense()
        return svd_truncated(x, cutoff, cutoff_mode, max_bond, absorb)

    U, s, V = sli.svd(x, k)
    VH = dag_numba(V)
    return _trim_and_renorm_svd_result_numba(
        U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


def _rsvd_numpy(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0, renorm=0):
    if max_bond > 0:
        if cutoff > 0.0:
            # adapt and block
            U, s, VH = rand_linalg.rsvd(x, cutoff, k_max=max_bond)
        else:
            U, s, VH = rand_linalg.rsvd(x, max_bond)
    else:
        U, s, VH = rand_linalg.rsvd(x, cutoff)

    return _trim_and_renorm_svd_result_numba(
        U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


def rsvd(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0, renorm=0):
    """SVD-decomposition using randomized methods (due to Halko). Allows the
    computation of only a certain number of singular values, e.g. max_bond,
    from the get-go, and is thus more efficient. Can also supply
    ``scipy.sparse.linalg.LinearOperator``.
    """
    if isinstance(x, (np.ndarray, spla.LinearOperator)):
        return _rsvd_numpy(x, cutoff, cutoff_mode, max_bond, absorb, renorm)

    U, s, VH = do("linalg.rsvd", x, max_bond)
    return _trim_and_renorm_svd_result(
        U, s, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


def eigsh(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0, renorm=0):
    """SVD-decomposition using iterative hermitian eigen decomp, thus assuming
    that ``x`` is hermitian. Allows the computation of only a certain number of
    singular values, e.g. max_bond, from the get-go, and is thus more
    efficient. Can also supply ``scipy.sparse.linalg.LinearOperator``.
    """
    k = _choose_k(x, cutoff, max_bond)

    if k == "full":
        if not isinstance(x, np.ndarray):
            x = x.to_dense()
        return eigh(x, cutoff, cutoff_mode, max_bond, absorb)

    s, U = base_linalg.eigh(x, k=k)
    s, U = s[::-1], U[:, ::-1]  # make sure largest singular value first
    V = ldmul_numba(sgn(s), dag_numba(U))
    s = np.abs(s)
    return _trim_and_renorm_svd_result_numba(
        U, s, V, cutoff, cutoff_mode, max_bond, absorb, renorm
    )


@compose
def qr_stabilized(x, backend=None):
    """QR-decomposition, with stabilized R factor."""
    with backend_like(backend):
        Q, R = do("linalg.qr", x)
        # stabilize the diagonal of R
        rd = do("diag", R)
        s = sgn(rd)
        Q = rdmul(Q, s)
        R = ldmul(s, R)
        return Q, None, R


@qr_stabilized.register("numpy")
@njit  # pragma: no cover
def qr_stabilized_numba(x):
    Q, R = np.linalg.qr(x)
    for i in range(R.shape[0]):
        rii = R[i, i]
        si = sgn_numba(rii)
        if si != 1.0:
            Q[:, i] *= si
            R[i, i:] *= si
    return Q, None, R


@compose
def lq_stabilized(x, backend=None):
    with backend_like(backend):
        Q, _, L = qr_stabilized(do("transpose", x))
        return do("transpose", L), None, do("transpose", Q)


@lq_stabilized.register("numpy")
@njit  # pragma: no cover
def lq_stabilized_numba(x):
    Q, _, L = qr_stabilized_numba(x.T)
    return L.T, None, Q.T



@njit  # pragma: no cover
def _cholesky_numba(x, cutoff=-1, cutoff_mode=3, max_bond=-1, absorb=0):
    """SVD-decomposition, using cholesky decomposition, only works if
    ``x`` is positive definite.
    """
    L = np.linalg.cholesky(x)
    return L, None, dag_numba(L)


def cholesky(x, cutoff=-1, cutoff_mode=3, max_bond=-1, absorb=0):
    try:
        return _cholesky_numba(x, cutoff, cutoff_mode, max_bond, absorb)
    except np.linalg.LinAlgError as e:
        if cutoff < 0:
            raise e
        # try adding cutoff identity - assuming it is approx allowable error
        xi = x + 2 * cutoff * np.eye(x.shape[0])
        return _cholesky_numba(xi, cutoff, cutoff_mode, max_bond, absorb)


# ------ similarity transforms for compressing effective environments ------- #


def _similarity_compress_eig(X, max_bond, renorm):
    # eigen decompose X -> V w V^-1
    el, ev = do("linalg.eig", X)
    evi = do("linalg.inv", ev)

    # choose largest abs value eigenpairs
    sel = do("argsort", do("abs", el))[-max_bond:]
    Cl = ev[:, sel]
    Cr = evi[sel, :]

    if renorm:
        trace_old = do("sum", el)
        trace_new = do("sum", el[sel])
        Cl = Cl * trace_old / trace_new

    return Cl, Cr


@njit(
    [
        "(c8[:,:], i4, i4)",
        "(c16[:,:], i4, i4)",
    ]
)  # pragma: no cover
def _similarity_compress_eig_numba(X, max_bond, renorm):
    el, ev = np.linalg.eig(X)
    evi = np.linalg.inv(ev)
    sel = np.argsort(np.abs(el))[-max_bond:]
    Cl = ev[:, sel]
    Cr = evi[sel, :]
    if renorm:
        trace_old = np.sum(el)
        trace_new = np.sum(el[sel])
        Cl = Cl * trace_old / trace_new
    return Cl, Cr


def _similarity_compress_eigh(X, max_bond, renorm):
    XX = (X + dag(X)) / 2
    el, ev = do("linalg.eigh", XX)
    sel = do("argsort", do("abs", el))[-max_bond:]
    Cl = ev[:, sel]
    Cr = dag(Cl)
    if renorm:
        trace_old = do("trace", X)
        trace_new = do("trace", Cr @ (X @ Cl))
        Cl = Cl * trace_old / trace_new
    return Cl, Cr


@njit  # pragma: no cover
def _similarity_compress_eigh_numba(X, max_bond, renorm):
    XX = (X + dag_numba(X)) / 2
    el, ev = np.linalg.eigh(XX)
    sel = np.argsort(-np.abs(el))[:max_bond]
    Cl = ev[:, sel]
    Cr = dag_numba(Cl)
    if renorm:
        trace_old = np.trace(X)
        trace_new = np.trace(Cr @ (X @ Cl))
        Cl = Cl * trace_old / trace_new
    return Cl, Cr


def _similarity_compress_svd(X, max_bond, renorm, asymm):
    U, _, VH = do("linalg.svd", X)
    U = U[:, :max_bond]

    Cl = U
    if asymm:
        VH = VH[:max_bond, :]
        Cr = dag(U)
        Cl = dag(VH)
    else:
        Cr = dag(U)

    if renorm:
        # explicitly maintain trace value
        trace_old = do("trace", X)
        trace_new = do("trace", Cr @ (X @ Cl))
        Cl = Cl * (trace_old / trace_new)

    return Cl, Cr


@njit  # pragma: no cover
def _similarity_compress_svd_numba(X, max_bond, renorm, asymm):
    U, _, VH = np.linalg.svd(X)
    U = U[:, :max_bond]
    Cl = U

    if asymm:
        VH = VH[:max_bond, :]
        Cr = dag_numba(U)
        Cl = dag_numba(VH)
    else:
        Cr = dag_numba(U)

    if renorm:
        trace_old = np.trace(X)
        trace_new = np.trace(Cr @ (X @ Cl))
        Cl = Cl * trace_old / trace_new
    return Cl, Cr


def _similarity_compress_biorthog(X, max_bond, renorm):
    U, s, VH = do("linalg.svd", X)

    B = U[:, :max_bond]
    AH = VH[:max_bond, :]

    Uab, sab, VHab = do("linalg.svd", AH @ B)
    sab = (sab + 1e-12 * do("max", sab)) ** -0.5
    sab_inv = do("reshape", sab, (1, -1))
    P = Uab * sab_inv
    Q = dag(VHab) * sab_inv

    Cl = B @ Q
    Cr = dag(P) @ AH

    if renorm:
        trace_old = do("trace", X)
        trace_new = do("trace", Cr @ (X @ Cl))
        Cl = Cl * trace_old / trace_new

    return Cl, Cr


@njit  # pragma: no cover
def _similarity_compress_biorthog_numba(X, max_bond, renorm):
    U, s, VH = np.linalg.svd(X)

    B = U[:, :max_bond]
    AH = VH[:max_bond, :]

    Uab, sab, VHab = np.linalg.svd(AH @ B)

    # smudge factor
    sab += 1e-12 * np.max(sab)
    sab **= -0.5

    sab_inv = sab.reshape((1, -1))
    P = Uab * sab_inv
    Q = dag_numba(VHab) * sab_inv

    Cl = B @ Q
    Cr = dag_numba(P) @ AH

    if renorm:
        trace_old = np.trace(X)
        trace_new = np.trace(Cr @ (X @ Cl))
        Cl = Cl * trace_old / trace_new

    return Cl, Cr


_similarity_compress_fns = {
    ("eig", False): _similarity_compress_eig,
    ("eig", True): _similarity_compress_eig_numba,
    ("eigh", False): _similarity_compress_eigh,
    ("eigh", True): _similarity_compress_eigh_numba,
    ("svd", False): functools.partial(_similarity_compress_svd, asymm=0),
    ("svd", True): functools.partial(_similarity_compress_svd_numba, asymm=0),
    ("biorthog", False): _similarity_compress_biorthog,
    ("biorthog", True): _similarity_compress_biorthog_numba,
}


def similarity_compress(X, max_bond, renorm=True, method="eigh"):
    if method == "eig":
        if get_dtype_name(X) == "float64":
            X = astype(X, "complex128")
        elif get_dtype_name(X) == "float32":
            X = astype(X, "complex64")

    isnumpy = isinstance(X, np.ndarray)
    # if isnumpy:
    #     X = np.ascontiguousarray(X)
    fn = _similarity_compress_fns[method, isnumpy]
    return fn(X, max_bond, int(renorm))


@compose
def isometrize_qr(x, backend=None):
    """Perform isometrization using the QR decomposition.
    """
    with backend_like(backend):
        fat = x.shape[0] < x.shape[1]
        if fat:
            x = do("transpose", x)
        Q, R = do('linalg.qr', x)
        # stabilize qr by fixing diagonal of R in canonical, positive form (we
        # don't actaully do anything to R, just absorb the necessary sign -> Q)
        rd = do('diag', R)
        s = do('sign', rd) + (rd == 0)
        Q = Q * reshape(s, (1, -1))
        if fat:
            Q = do("transpose", Q)
        return Q


@compose
def isometrize_svd(x, backend=None):
    with backend_like(backend):
        fat = x.shape[0] < x.shape[1]
        if fat:
            x = do("transpose", x)

        Q = do('linalg.svd', x)[0]
        if fat:
            Q = do("transpose", Q)

        return Q


@compose
def isometrize_exp(x, backend):
    r"""Perform isometrization using the using anti-symmetric matrix
    exponentiation.

    .. math::

            U_A = \exp{A - A^\dagger}

    If ``x`` is rectangular it is completed with zeros first.
    """
    with backend_like(backend):
        m, n = x.shape
        d = max(m, n)
        x = do(
            'pad', x, [[0, d - m], [0, d - n]], 'constant', constant_values=0.0
        )
        expx = do('linalg.expm', x - dag(x))
        return expx[:m, :n]


@compose
def isometrize_modified_gram_schmidt(A, backend=None):
    """Perform isometrization explicitly using the modified Gram Schmidt
    procedure.
    """
    with backend_like(backend):
        m, n = A.shape

        thin = m > n
        if thin:
            A = do('transpose', A)

        Q = []
        for j in range(0, min(m, n)):

            q = A[j, :]
            for i in range(0, j):
                rij = do('tensordot', do('conj', Q[i]), q, 1)
                q = q - rij * Q[i]

            Q.append(q / do('linalg.norm', q, 2))

        Q = do('stack', tuple(Q), axis=0, like=A)

        if thin:
            Q = do('transpose', Q)

        return Q


_ISOMETRIZE_METHODS = {
    'qr': isometrize_qr,
    'svd': isometrize_svd,
    'exp': isometrize_exp,
    'mgs': isometrize_modified_gram_schmidt,
}


def isometrize(x, method='qr'):
    """Generate a isometric (or unitary if square) matrix from array ``x``.

    Parameters
    ----------
    x : array
        The matrix to project into isometrix form.
    method : {'qr', 'svd', 'exp', 'mgs'}, optional
        The method used to generate the isometry. Note ``'qr'`` is the fastest
        and most robust but, for example, some libraries cannot back-propagate
        through it.
    """
    return _ISOMETRIZE_METHODS[method](x)



import pytest
import operator
import importlib

import numpy as np
from numpy.testing import assert_allclose
import scipy.sparse.linalg as spla

import quimb as qu
import quimb.tensor as qtn
from quimb.tensor import (
    oset,
    bonds,
    tensor_contract,
    tensor_direct_product,
    Tensor,
    TensorNetwork,
    TensorNetwork1D,
    rand_tensor,
    MPS_rand_state,
    TNLinearOperator1D,
)
from quimb.tensor.decomp import _compute_number_svals_to_keep_numba
from quimb.tensor.contraction import _CONTRACT_BACKEND, _TENSOR_LINOP_BACKEND

autograd_mark = pytest.mark.skipif(
    importlib.util.find_spec('autograd') is None,
    reason='autograd not installed')


def test_trim_singular_vals():
    s = np.array([3., 2., 1., 0.1])
    assert _compute_number_svals_to_keep_numba(s, 0.5, 1) == 3
    assert _compute_number_svals_to_keep_numba(s, 0.5, 2) == 2
    assert _compute_number_svals_to_keep_numba(s, 2, 3) == 2
    assert _compute_number_svals_to_keep_numba(s, 5.02, 3) == 1


class TestContractOpts:

    def test_contract_strategy(self):
        assert qtn.get_contract_strategy() == 'greedy'
        with qtn.contract_strategy('auto'):
            assert qtn.get_contract_strategy() == 'auto'
        assert qtn.get_contract_strategy() == 'greedy'

    def test_contract_backend(self):
        assert qtn.get_contract_backend() == _CONTRACT_BACKEND
        with qtn.contract_backend('cupy'):
            assert qtn.get_contract_backend() == 'cupy'
        assert qtn.get_contract_backend() == _CONTRACT_BACKEND

    def test_tensor_linop_backend(self):
        assert qtn.get_tensor_linop_backend() == _TENSOR_LINOP_BACKEND
        with qtn.tensor_linop_backend('cupy'):
            assert qtn.get_tensor_linop_backend() == 'cupy'
        assert qtn.get_tensor_linop_backend() == _TENSOR_LINOP_BACKEND

    def test_contract_cache(self):
        import tempfile
        import os
        from opt_einsum.paths import register_path_fn

        info = {'num_calls': 0}

        def my_custom_opt(inputs, output, size_dict, memory_limit=None):
            info['num_calls'] += 1
            return [(0, 1)] * (len(inputs) - 1)

        register_path_fn('quimb_test_opt', my_custom_opt)

        tn = qtn.MPS_rand_state(4, 3) & qtn.MPS_rand_state(4, 3)
        assert (
            tn.contract(all, optimize='quimb_test_opt', get='expression')
            is
            tn.contract(all, optimize='quimb_test_opt', get='expression'))
        assert info['num_calls'] == 1

        # contraction pathinfo objects are now cached together
        assert (
            tn.contract(all, optimize='quimb_test_opt', get='path-info')
            is
            tn.contract(all, optimize='quimb_test_opt', get='path-info'))
        assert info['num_calls'] == 1

        # set a directory cache - functions will be run fresh again
        with tempfile.TemporaryDirectory() as tdir:
            assert len(os.listdir(tdir)) == 0
            qtn.set_contract_path_cache(tdir)
            assert (
                tn.contract(all, optimize='quimb_test_opt', get='expression')
                is
                tn.contract(all, optimize='quimb_test_opt', get='expression'))
            assert info['num_calls'] == 2
            assert (
                tn.contract(all, optimize='quimb_test_opt', get='path-info')
                is
                tn.contract(all, optimize='quimb_test_opt', get='path-info'))
            assert info['num_calls'] == 2
            assert len(os.listdir(tdir)) != 0

            # need to release close the cache so the directory can be deleted
            qtn.set_contract_path_cache(None)


class TestBasicTensorOperations:

    def test_tensor_construct(self):
        x = np.random.randn(2, 3, 4)
        a = Tensor(x, inds=[0, 1, 2], tags='blue')
        assert_allclose(a.H.data, x.conj())
        assert a.size == 24

        with pytest.raises(ValueError):
            Tensor(x, inds=[0, 2], tags='blue')

        assert repr(a) == ("Tensor(shape=(2, 3, 4), "
                           "inds=(0, 1, 2), tags=oset(['blue']))")
        assert str(a) == ("Tensor(shape=(2, 3, 4), inds=(0, 1, 2), "
                          "tags=oset(['blue']), backend='numpy', "
                          "dtype='float64')")

    def test_tensor_copy(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2], tags='blue')
        b = a.copy()
        b.add_tag('foo')
        assert 'foo' not in a.tags
        b.data[:] = b.data / 2
        # still reference the same underlying array
        assert_allclose(a.data, b.data)

    def test_tensor_deep_copy(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2], tags='blue')
        b = a.copy(deep=True)
        b.add_tag('foo')
        assert 'foo' not in a.tags
        b.data[:] = b.data / 2
        # still reference the same underlying array
        assert_allclose(a.data / 2, b.data)

    def test_with_alpha_construct(self):
        x = np.random.randn(2, 3, 4)
        a = Tensor(x, inds='ijk', tags='blue')
        assert_allclose(a.H.data, x.conj())
        assert a.size == 24

        with pytest.raises(ValueError):
            Tensor(x, inds='ij', tags='blue')

        x = np.random.randn(2, 3, 4)
        a = Tensor(x, inds=['a1', 'b2', 'c3'], tags='blue')
        assert_allclose(a.H.data, x.conj())
        assert a.size == 24

        with pytest.raises(ValueError):
            Tensor(x, inds=['ijk'], tags='blue')

    def test_arithmetic_scalar(self):
        x = np.random.randn(2, 3, 4)
        a = Tensor(x, inds=[0, 1, 2], tags='blue')
        assert_allclose((a + 2).data, x + 2)
        assert_allclose((a - 3).data, x - 3)
        assert_allclose((a * 4).data, x * 4)
        assert_allclose((a / 5).data, x / 5)
        assert_allclose((a ** 2).data, x ** 2)
        assert_allclose((2 + a).data, 2 + x)
        assert_allclose((3 - a).data, 3 - x)
        assert_allclose((4 * a).data, 4 * x)
        assert_allclose((5 / a).data, 5 / x)
        assert_allclose((5 ** a).data, 5 ** x)

    @pytest.mark.parametrize("op", [operator.__add__,
                                    operator.__sub__,
                                    operator.__mul__,
                                    operator.__pow__,
                                    operator.__truediv__])
    @pytest.mark.parametrize("mismatch", (True, False))
    def test_tensor_tensor_arithmetic(self, op, mismatch):
        a = Tensor(np.random.rand(2, 3, 4), inds=[0, 1, 2], tags='blue')
        b = Tensor(np.random.rand(2, 3, 4), inds=[0, 1, 2], tags='red')
        if mismatch:
            b.modify(inds=(0, 1, 3))
            with pytest.raises(ValueError):
                op(a, b)
        else:
            c = op(a, b)
            assert_allclose(c.data, op(a.data, b.data))

    def test_tensor_conj_inplace(self):
        data = np.random.rand(2, 3, 4) + 1.0j * np.random.rand(2, 3, 4)
        a = Tensor(data, inds=[0, 1, 2], tags='blue')
        a.conj_()
        assert_allclose(data.conj(), a.data)

    def test_contract_some(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2])
        b = Tensor(np.random.randn(3, 4, 5), inds=[1, 2, 3])

        assert a.shared_bond_size(b) == 12

        c = a @ b

        assert isinstance(c, Tensor)
        assert c.shape == (2, 5)
        assert c.inds == (0, 3)

    def test_contract_all(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2])
        b = Tensor(np.random.randn(3, 4, 2), inds=[1, 2, 0])
        c = a @ b
        assert isinstance(c, float)
        assert not isinstance(c, Tensor)

    def test_contract_None(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2])
        b = Tensor(np.random.randn(3, 4, 5), inds=[3, 4, 5])
        c = a @ b
        assert c.shape == (2, 3, 4, 3, 4, 5)
        assert c.inds == (0, 1, 2, 3, 4, 5)

        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2])
        b = Tensor(np.random.randn(3, 4, 5), inds=[5, 4, 3])
        c = a @ b

        assert c.shape == (2, 3, 4, 3, 4, 5)
        assert c.inds == (0, 1, 2, 5, 4, 3)

    def test_raise_on_triple_inds(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2])
        b = Tensor(np.random.randn(3, 4, 5), inds=[1, 1, 2])
        with pytest.raises(ValueError):
            a @ b

    def test_multi_contract(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2],
                   tags='red')
        b = Tensor(np.random.randn(3, 4, 5), inds=[1, 2, 3],
                   tags='blue')
        c = Tensor(np.random.randn(5, 2, 6), inds=[3, 0, 4],
                   tags='blue')
        d = tensor_contract(a, b, c)
        assert isinstance(d, Tensor)
        assert d.shape == (6,)
        assert d.inds == (4,)
        assert d.tags == oset(('red', 'blue'))

    def test_contract_with_legal_characters(self):
        a = Tensor(np.random.randn(2, 3, 4), inds='abc',
                   tags='red')
        b = Tensor(np.random.randn(3, 4, 5), inds='bcd',
                   tags='blue')
        c = a @ b
        assert c.shape == (2, 5)
        assert c.inds == ('a', 'd')

    def test_contract_with_out_of_range_inds(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[-1, 100, 2200],
                   tags='red')
        b = Tensor(np.random.randn(3, 4, 5), inds=[100, 2200, -3],
                   tags='blue')
        c = a @ b
        assert c.shape == (2, 5)
        assert c.inds == (-1, -3)

    def test_contract_with_wild_mix(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=['-1', 'a', 'foo'],
                   tags='red')
        b = Tensor(np.random.randn(3, 4, 5), inds=['a', 'foo', '42.42'],
                   tags='blue')
        c = a @ b
        assert c.shape == (2, 5)
        assert c.inds == ('-1', '42.42')

    def test_fuse(self):
        a = Tensor(np.random.rand(2, 3, 4, 5), 'abcd', tags={'blue'})
        b = a.fuse({'bra': ['a', 'c'], 'ket': 'bd'})
        assert set(b.shape) == {8, 15}
        assert set(b.inds) == {'bra', 'ket'}
        assert b.tags == oset(('blue',))

        b = a.fuse({'ket': 'bd', 'bra': 'ac'})
        assert set(b.shape) == {15, 8}
        assert set(b.inds) == {'ket', 'bra'}
        assert b.tags == oset(('blue',))

    def test_unfuse(self):
        a = Tensor(np.random.rand(2, 3, 4, 5), 'abcd', tags={'blue'})
        b = a.fuse({'bra': ['a', 'c'], 'ket': 'bd'})

        c = b.unfuse({'bra': ['a', 'c'], 'ket': 'bd'},
                     {'bra': [2, 4], 'ket': [3, 5]})
        assert set(c.shape) == {2, 3, 4, 5}
        assert set(c.inds) == {'a', 'b', 'c', 'd'}
        assert c.left_inds == b.left_inds
        assert np.allclose(c.data.reshape(8, 15), b.data)

        b.modify(left_inds=['ket'])
        c = b.unfuse({'ket': 'bd'}, {'ket': [5, 3]})
        assert set(c.shape) == {3, 5, 8}
        assert set(c.inds) == {'b', 'd', 'bra'}
        assert set(c.tags) == {'blue'}
        assert set(c.left_inds) == {'b', 'd'}

        b.modify(left_inds=['bra'])
        c = b.unfuse({'ket': 'bd'}, {'ket': [5, 3]})
        assert set(c.left_inds) == {'bra'}

    def test_fuse_leftover(self):
        a = Tensor(np.random.rand(2, 3, 4, 5, 2, 2), 'abcdef', tags={'blue'})
        b = a.fuse({'bra': 'ac', 'ket': 'bd'})
        assert b.shape == (8, 15, 2, 2)
        assert b.inds == ('bra', 'ket', 'e', 'f')
        assert b.tags == oset(('blue',))

    def test_tensor_transpose(self):
        a = Tensor(np.random.rand(2, 3, 4, 5, 2, 2), 'abcdef', tags={'blue'})
        at = a.transpose(*'cdfeba')
        assert at.shape == (4, 5, 2, 2, 3, 2)
        assert at.inds == ('c', 'd', 'f', 'e', 'b', 'a')

        with pytest.raises(ValueError):
            a.transpose(*'cdfebz')

    def test_tensor_moveindex(self):
        A = rand_tensor([3, 4, 5], ['a', 'b', 'c'])
        B = rand_tensor([3, 4, 5], ['a', 'b', 'c'])
        x = A @ B
        A.moveindex_('b', 0)
        B.moveindex_('b', -1)
        assert A.inds[0] == 'b' and B.inds[2] == 'b'
        assert A @ B == pytest.approx(x)

    def test_tensor_trace(self):
        t = qtn.rand_tensor((3, 3, 3), 'abc', dtype='complex128')
        tb = t.trace('a', 'c')
        assert tb.inds == ('b',)
        assert_allclose(tb.data, np.trace(t.data, axis1=0, axis2=2))
        tc = t.trace('a', 'b')
        assert tc.inds == ('c',)
        assert_allclose(tc.data, np.trace(t.data, axis1=0, axis2=1))
        with pytest.raises(ValueError):
            t.trace('a', 'z')
        assert not isinstance(
            qtn.rand_tensor([2, 2], 'ab').trace('a', 'b'),
            qtn.Tensor)
        assert isinstance(
            qtn.rand_tensor([2, 2], 'ab').trace('a', 'b',
                                                preserve_tensor=True),
            qtn.Tensor)

    def test_tensor_trace_multi(self):
        t = qtn.rand_tensor((3, 3, 3, 3, 3), 'abcde', dtype='complex128')
        t1 = t.trace(['a', 'c'], ['e', 'b'])
        te = t.trace('a', 'e').trace('c', 'b')
        assert t1.almost_equals(te)
        with pytest.raises(ValueError):
            t.trace(['a', 'b', 'c'], ['d', 'e'])

    def test_sum_reduce(self):
        t = rand_tensor((2, 3, 4), 'abc')
        ta = t.sum_reduce('a')
        assert ta.ndim == 2
        assert_allclose(ta.data, t.data.sum(axis=0))
        tb = t.sum_reduce('b')
        assert tb.ndim == 2
        assert_allclose(tb.data, t.data.sum(axis=1))
        tc = t.sum_reduce('c')
        assert tc.ndim == 2
        assert_allclose(tc.data, t.data.sum(axis=2))
        with pytest.raises(ValueError):
            t.sum_reduce_('d')

    def test_ownership(self):
        a = rand_tensor((2, 2), ('a', 'b'), tags={'X', 'Y'})
        b = rand_tensor((2, 2), ('b', 'c'), tags={'X', 'Z'})
        assert not a.check_owners()
        assert not b.check_owners()
        tn = TensorNetwork((a, b), virtual=True)
        assert a.check_owners()
        assert b.check_owners()
        assert a.owners[hash(tn)][0]() is tn
        assert b.owners[hash(tn)][0]() is tn
        assert all(map(tn.ind_map.__contains__, ('a', 'b', 'c')))
        assert all(map(tn.tag_map.__contains__, ('X', 'Y', 'Z')))
        a.reindex_({'a': 'd'})
        assert 'a' not in tn.ind_map
        assert 'd' in tn.ind_map
        assert len(tn.tag_map['X']) == 2
        b.retag_({'X': 'W'})
        assert len(tn.tag_map['X']) == 1
        assert 'W' in tn.tag_map
        del tn
        assert not a.check_owners()
        assert not b.check_owners()

    def test_isel(self):
        T = rand_tensor((2, 3, 4, 5, 6), inds=['a', 'b', 'c', 'd', 'e'])
        Tis = T.isel({'d': 2, 'b': 0})
        assert Tis.shape == (2, 4, 6)
        assert Tis.inds == ('a', 'c', 'e')
        assert_allclose(Tis.data, T.data[:, 0, :, 2, :])

    def test_cut_iter(self):
        psi = MPS_rand_state(10, 7, cyclic=True)
        pp = psi.H & psi
        bnds = bonds(pp[0], pp[-1])
        assert sum(tn ^ all for tn in pp.cut_iter(*bnds)) == pytest.approx(1.0)
        assert pp ^ all == pytest.approx(1.0)

    @pytest.mark.parametrize("method", ['qr', 'exp', 'mgs', 'svd'])
    def test_isometrize(self, method):
        t = rand_tensor((2, 3, 4), 'abc')
        assert t.H @ t != pytest.approx(3.0)
        t.isometrize('b', inplace=True, method=method)
        assert t.H @ t == pytest.approx(3.0)
        assert t.inds == ('b', 'a', 'c')

    def test_connect(self):
        x = rand_tensor((2, 3), 'ab')
        y = rand_tensor((3, 2), 'cd')

        with pytest.raises(ValueError):
            qtn.connect(x, y, 0, 0)

        tn = x | y
        assert len(tn.outer_inds()) == 4
        qtn.connect(x, y, 0, 1)
        assert len(tn.outer_inds()) == 2
        qtn.connect(x, y, 1, 0)
        assert len(tn.outer_inds()) == 0
        assert tn.contract(all, preserve_tensor=True).shape == ()
        # make sure bond is newly labelled
        assert set('abcd') & set(tn.all_inds()) == set()

    def test_group_inds(self):
        x = rand_tensor((2, 2, 2, 2), 'abcd')
        y = rand_tensor((2, 2, 2), 'bdf')
        lix, six, rix = qtn.group_inds(x, y)
        assert lix == ['a', 'c']
        assert six == ['b', 'd']
        assert rix == ['f']

    def test_group_inds_tensor_network(self):
        tn = qtn.TN2D_with_value(1.0, 4, 4, 2)
        ltn = tn.select(['I1,1', 'I1,2'], 'any')
        rtn = tn.select(['I2,1', 'I2,2'], 'any')
        lix, six, rix = qtn.group_inds(ltn, rtn)
        assert len(lix) == len(rix) == 4
        assert len(six) == 2
        assert ltn.inds_size(six) == 4


class TestTensorFunctions:
    @pytest.mark.parametrize('method', ['svd', 'eig', 'isvd', 'svds'])
    @pytest.mark.parametrize('linds', [('a', 'b', 'd'), ('c', 'e')])
    @pytest.mark.parametrize('cutoff', [-1.0, 1e-13, 1e-10])
    @pytest.mark.parametrize('cutoff_mode', ['abs', 'rel', 'sum2'])
    @pytest.mark.parametrize('absorb', ['left', 'both', 'right'])
    def test_split_tensor_rank_revealing(self, method, linds, cutoff,
                                         cutoff_mode, absorb):
        a = rand_tensor((2, 3, 4, 5, 6), inds='abcde', tags='red')
        a_split = a.split(linds, method=method, cutoff=cutoff,
                          cutoff_mode=cutoff_mode, absorb=absorb)
        assert len(a_split.tensors) == 2
        if linds == 'abd':
            assert ((a_split.shape == (2, 3, 5, 4, 6)) or
                    (a_split.shape == (4, 6, 2, 3, 5)))
        elif linds == 'edc':
            assert ((a_split.shape == (6, 5, 4, 2, 3)) or
                    (a_split.shape == (2, 3, 6, 5, 4)))
        assert (a_split ^ ...).almost_equals(a)

    @pytest.mark.parametrize('method', ['qr', 'lq'])
    @pytest.mark.parametrize('linds', [('a', 'b', 'd'), ('c', 'e')])
    def test_split_tensor_rank_hidden(self, method, linds):
        a = rand_tensor((2, 3, 4, 5, 6), inds='abcde', tags='red')
        a_split = a.split(linds, method=method)
        assert len(a_split.tensors) == 2
        if linds == 'abd':
            assert ((a_split.shape == (2, 3, 5, 4, 6)) or
                    (a_split.shape == (4, 6, 2, 3, 5)))
        elif linds == 'edc':
            assert ((a_split.shape == (6, 5, 4, 2, 3)) or
                    (a_split.shape == (2, 3, 6, 5, 4)))
        assert (a_split ^ ...).almost_equals(a)

    @pytest.mark.parametrize('method', ['svd', 'eig'])
    def test_singular_values(self, method):
        psim = Tensor(np.eye(2) * 2**-0.5, inds='ab')
        assert_allclose(psim.H @ psim, 1.0)
        assert_allclose(psim.singular_values('a', method=method)**2,
                        [0.5, 0.5])

    def test_split_renorm(self):
        t = rand_tensor((3, 3, 3, 3), ['a', 'b', 'c', 'd'])
        n_nuc = t.singular_values(['a', 'b']).sum()
        n_fro = (t.singular_values(['a', 'b'])**2).sum()**0.5

        tc = t.split(['a', 'b'], cutoff=0.0, max_bond=5, renorm=1) ^ all
        nc_nuc = tc.singular_values(['a', 'b']).sum()
        nc_fro = (tc.singular_values(['a', 'b'])**2).sum()**0.5
        assert nc_nuc == pytest.approx(n_nuc)
        assert nc_fro != pytest.approx(n_fro)

        tc = t.split(['a', 'b'], cutoff=0.0, max_bond=5, renorm=2) ^ all
        nc_nuc = tc.singular_values(['a', 'b']).sum()
        nc_fro = (tc.singular_values(['a', 'b'])**2).sum()**0.5
        assert nc_fro == pytest.approx(n_fro)
        assert nc_nuc != pytest.approx(n_nuc)

    def test_absorb_none(self):
        x = qtn.rand_tensor((4, 5, 6, 7), inds='abcd', tags='X', seed=42)
        e = x.H @ x

        with pytest.raises(ValueError):
            x.split(['a', 'c'], absorb=None, method='qr')

        xs_tn = x.split(['a', 'c'], absorb=None, stags='S')
        assert isinstance(xs_tn, TensorNetwork)
        assert xs_tn.num_tensors == 3
        e1 = (xs_tn.H & xs_tn).contract(all, output_inds=())
        assert e1 == pytest.approx(e)
        assert 'S' in xs_tn.tags

        Tl, Ts, Tr = x.split(['a', 'c'], absorb=None, get='tensors')
        assert isinstance(Ts, Tensor)
        assert len(Ts.inds) == 1
        assert 'X' in Ts.tags
        Tl.multiply_index_diagonal_(Ts.inds[0], Ts.data)
        xs_tn = Tl & Tr
        e2 = (xs_tn.H & xs_tn).contract(all)
        assert e2 == pytest.approx(e)

        l, s, r = x.split(['a', 'c'], absorb=None, get='arrays')
        assert s.size == 24
        y_data = np.einsum('acx,x,xbd->abcd', l, s, r)
        assert_allclose(y_data, x.data)

        l, s, r = x.split(['a', 'c'], absorb=None, get='arrays', max_bond=20)
        assert s.size == 20
        y_data = np.einsum('acx,x,xbd->abcd', l, s, r)
        assert (
            qu.norm(y_data, 'fro') ==
            pytest.approx(qu.norm(x.data, 'fro'), rel=0.1)
        )

    @pytest.mark.parametrize('method', ['svd', 'eig'])
    def test_renorm(self, method):
        U = qu.rand_uni(10)
        s = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        x = np.einsum('ab,b,bc->ac', U, s, qu.dag(U))

        t = qtn.Tensor(x, inds='ab')
        fn2 = t.norm()**2
        trc = np.einsum('aa', t.data).real

        assert fn2 == pytest.approx(385.)
        assert trc == pytest.approx(55.)

        tn2 = t.split('a', method='svd', cutoff=0.1, cutoff_mode='rsum2')
        a_fn2 = tn2.H @ tn2
        assert qtn.bonds_size(*tn2) == 6
        assert a_fn2 == pytest.approx(fn2)

        tn2 = t.split('a', method='svd', cutoff=40, cutoff_mode='sum2')
        a_fn2 = tn2.H @ tn2
        assert qtn.bonds_size(*tn2) == 6
        assert a_fn2 == pytest.approx(fn2)

        tn1 = t.split('a', method='svd', cutoff=0.2, cutoff_mode='rsum1')
        a_trc = tn1.trace('a', 'b').real
        assert qtn.bonds_size(*tn1) == 6
        assert a_trc == pytest.approx(trc)

        tn1 = t.split('a', method='svd', cutoff=11, cutoff_mode='sum1')
        a_trc = tn1.trace('a', 'b').real
        assert qtn.bonds_size(*tn1) == 6
        assert a_trc == pytest.approx(trc)

    @pytest.mark.parametrize('method', ['svd', 'eig'])
    def test_entropy(self, method):
        psim = Tensor(np.eye(2) * 2**-0.5, inds='ab')
        assert_allclose(psim.H @ psim, 1.0)
        assert_allclose(psim.entropy('a', method=method)**2, 1)

    @pytest.mark.parametrize('method', ['svd', 'eig'])
    def test_entropy_matches_dense(self, method):
        p = MPS_rand_state(5, 32)
        p_dense = p.to_dense()
        real_svn = qu.entropy(p_dense.ptr([2] * 5, [0, 1, 2]))

        svn = (p ^ ...).entropy(('k0', 'k1', 'k2'))
        assert_allclose(real_svn, svn)

        # use tensor to left of bipartition
        p.canonize(2)
        t1 = p['I2']
        left_inds = set(t1.inds) - set(p['I3'].inds)
        svn = (t1).entropy(left_inds, method=method)
        assert_allclose(real_svn, svn)

        # use tensor to right of bipartition
        p.canonize(3)
        t2 = p['I3']
        left_inds = set(t2.inds) & set(p['I2'].inds)
        svn = (t2).entropy(left_inds, method=method)
        assert_allclose(real_svn, svn)

    def test_direct_product(self):
        a1 = rand_tensor((2, 3, 4), inds='abc')
        b1 = rand_tensor((3, 4, 5), inds='bcd')
        a2 = rand_tensor((2, 3, 4), inds='abc')
        b2 = rand_tensor((3, 4, 5), inds='bcd')

        c1 = (a1 @ b1) + (a2 @ b2)
        c2 = (tensor_direct_product(a1, a2, sum_inds=('a')) @
              tensor_direct_product(b1, b2, sum_inds=('d')))
        assert c1.almost_equals(c2)

    def test_direct_product_triple(self):
        a1 = rand_tensor((2, 3, 4), inds='abc')
        b1 = rand_tensor((3, 4, 5, 6), inds='bcde')
        c1 = rand_tensor((6, 7), inds='ef')

        a2 = rand_tensor((2, 3, 4), inds='abc')
        b2 = rand_tensor((3, 4, 5, 6), inds='bcde').transpose(*'decb')
        c2 = rand_tensor((6, 7), inds='ef')

        d1 = (a1 @ b1 @ c1) + (a2 @ b2 @ c2)
        d2 = (tensor_direct_product(a1, a2, sum_inds=('a')) @
              tensor_direct_product(b1, b2, sum_inds=('d')) @
              tensor_direct_product(c1, c2, sum_inds=('f')))
        assert d1.almost_equals(d2)

    @pytest.mark.parametrize("dtype", [float, complex, np.complex128,
                                       np.float_, np.float32, 'raise'])
    def test_rand_tensor(self, dtype):
        if dtype == 'raise':
            with pytest.raises(TypeError):
                rand_tensor((2, 3, 4), 'abc', dtype=dtype)
        else:
            t = rand_tensor((2, 3, 4), 'abc', dtype=dtype)
            assert t.dtype == np.dtype(dtype)

            tn = t & t
            assert tn.dtype == np.dtype(dtype)

    def test_squeeze(self):
        a = rand_tensor((1, 2, 3, 1, 4), inds='abcde', tags=['hello'])
        b = a.squeeze()
        assert b.shape == (2, 3, 4)
        assert b.inds == ('b', 'c', 'e')
        assert 'hello' in b.tags
        assert a.shape == (1, 2, 3, 1, 4)
        c = a.squeeze(include=['d'])
        assert c.shape == (1, 2, 3, 4)
        assert c.inds == ('a', 'b', 'c', 'e')

    def test_tensor_fuse_squeeze(self):
        a = rand_tensor((1, 2, 3), inds='abc')
        b = rand_tensor((2, 3, 4), inds='bcd')
        qtn.tensor_fuse_squeeze(a, b)
        assert a.inds == ('a', 'b')
        assert a.shape == (1, 6)
        assert b.inds == ('b', 'd')
        assert b.shape == (6, 4)

        a = rand_tensor((1, 1, 1), inds='abc')
        b = rand_tensor((1, 1, 1), inds='bcd')
        qtn.tensor_fuse_squeeze(a, b)
        assert a.inds == ('a',)
        assert a.shape == (1,)
        assert b.inds == ('d',)
        assert b.shape == (1,)

    @pytest.mark.parametrize('dtype', [None, 'complex128', 'float32'])
    def test_randomize(self, dtype):
        a = rand_tensor((2, 3, 4), ['a', 'b', 'c'], dtype='float64')
        if dtype is not None:
            assert a.dtype != dtype
        x1 = a.norm()
        a.randomize_(dtype=dtype)
        x2 = a.norm()
        assert x1 != pytest.approx(x2)
        assert a.shape == (2, 3, 4)
        if dtype is not None:
            assert a.dtype == dtype
        else:
            assert a.dtype == 'float64'

    def test_multiply_index_diagonal(self):
        x = rand_tensor((3, 4), 'ab')
        y = rand_tensor((4, 5), 'bc')
        z1 = x @ y
        # insert a diagonal gauge
        s = qu.randn(4)
        z2 = (
            x.multiply_index_diagonal('b', s) @
            y.multiply_index_diagonal('b', 1 / s)
        )
        assert z1.almost_equals(z2)

    @pytest.mark.parametrize("smudge", [1e-6, 1e-12])
    def test_balance_bonds(self, smudge):
        t1 = rand_tensor((3, 4), 'ab')
        t2 = rand_tensor((4, 5), 'bc')
        col_nrm_x1 = tensor_contract(t1.H, t1, output_inds='b').data
        col_nrm_y1 = tensor_contract(t2.H, t2, output_inds='b').data
        assert not np.allclose(col_nrm_x1, col_nrm_y1, rtol=1e-6)
        z1 = (t1 @ t2).data
        qtn.tensor_balance_bond(t1, t2, smudge=smudge)
        col_nrm_x2 = tensor_contract(t1.H, t1, output_inds='b').data
        col_nrm_y2 = tensor_contract(t2.H, t2, output_inds='b').data
        assert_allclose(col_nrm_x2, col_nrm_y2, rtol=10 * smudge)
        z2 = (t1 @ t2).data
        assert_allclose(z1, z2)

    def test_new_ind_with_identity(self):
        t = rand_tensor((2, 2, 3, 3), 'abcd')
        t.new_ind_with_identity(
            'switch', ['a', 'c'], ['b', 'd'], axis=2
        )
        assert t.inds == ('a', 'b', 'switch', 'c', 'd')
        assert t.isel({'switch': 1}).data.sum() == pytest.approx(6)


class TestTensorNetwork:
    def test_combining_tensors(self):
        a = rand_tensor((2, 3, 4), inds=[0, 1, 2], tags='red')
        b = rand_tensor((3, 4, 5), inds=[1, 2, 3], tags='blue')
        c = rand_tensor((5, 2, 6), inds=[3, 0, 4], tags='blue')

        with pytest.raises(TypeError):
            a & np.array([0, 0])

        abc1 = (a & b & c).H.contract()
        abc2 = (a & (b & c)).H.contract()
        abc3 = (TensorNetwork([a, b, c])).H.contract()
        abc4 = (TensorNetwork([a, TensorNetwork([b, c])])).H.contract()
        abc5 = (TensorNetwork([a]) & TensorNetwork([b, c])).H.contract()

        assert_allclose(abc1.data, abc2.data)
        assert_allclose(abc1.data, abc3.data)
        assert_allclose(abc1.data, abc4.data)
        assert_allclose(abc1.data, abc5.data)

    def test_copy(self):
        a = rand_tensor((2, 3, 4), inds='abc', tags='t0')
        b = rand_tensor((2, 3, 4), inds='abd', tags='t1')
        tn1 = TensorNetwork((a, b))
        tn2 = tn1.copy()
        # check can modify tensor structure
        tn2['t1'].modify(inds=('a', 'b', 'X'))
        assert tn1['t1'] is not tn2['t1']
        assert tn2['t1'].inds == ('a', 'b', 'X')
        assert tn1['t1'].inds == ('a', 'b', 'd')
        # but that data remains the same
        assert tn1['t1'].data is tn2['t1'].data
        tn2['t1'].data[:] /= 2
        assert_allclose(tn1['t1'].data, tn2['t1'].data)

    def test_copy_deep(self):
        a = rand_tensor((2, 3, 4), inds='abc', tags='t0')
        b = rand_tensor((2, 3, 4), inds='abd', tags='t1')
        tn1 = TensorNetwork((a, b))
        tn2 = tn1.copy(deep=True)
        # check can modify tensor structure
        tn2['t1'].modify(inds=('a', 'b', 'X'))
        assert tn1['t1'] is not tn2['t1']
        assert tn2['t1'].inds == ('a', 'b', 'X')
        assert tn1['t1'].inds == ('a', 'b', 'd')
        # and that data is not the same
        assert tn1['t1'].data is not tn2['t1'].data
        tn2['t1'].data[:] /= 2
        assert_allclose(tn1['t1'].data / 2, tn2['t1'].data)

    def test_TensorNetwork_init_checks(self):
        a = rand_tensor((2, 3, 4), inds=[0, 1, 2], tags={'red'})
        b = rand_tensor((3, 4, 5), inds=[1, 2, 3], tags={'blue'})
        c = rand_tensor((3, 4, 5), inds=[1, 2, 3], tags={'blue', 'c'})

        with pytest.raises(TypeError):
            TensorNetwork(a, b)  # missing brackets around ``a, b``.

        tn = a & b
        with pytest.raises(TypeError):
            tn['red'] = 1

        tn.add_tag('foo')
        assert len(tn['foo']) == 2
        with pytest.raises(KeyError):
            tn['foo'] = c

        tn[('foo', 'blue')] = c
        assert 'c' in tn.tags
        assert tn[('blue', 'c')] is c

        assert 'red' in tn.tags
        del tn['red']
        assert 'red' not in tn.tags

        assert set(tn.tag_map.keys()) == {'blue', 'c'}

        tn.drop_tags('c')
        assert set(tn.tag_map.keys()) == {'blue'}
        tn.drop_tags(['blue'])
        assert set(tn.tag_map.keys()) == set()

    def test_conj(self):
        a_data = np.random.randn(2, 3, 4) + 1.0j * np.random.randn(2, 3, 4)
        b_data = np.random.randn(3, 4, 5) + 1.0j * np.random.randn(3, 4, 5)
        c_data = np.random.randn(5, 2, 6) + 1.0j * np.random.randn(5, 2, 6)

        a = Tensor(a_data, inds=[0, 1, 2], tags={'red', '0'})
        b = Tensor(b_data, inds=[1, 2, 3], tags={'blue', '1'})
        c = Tensor(c_data, inds=[3, 0, 4], tags={'blue', '2'})

        tn = a & b & c
        new_tn = tn.conj()

        for i, arr in enumerate((a_data, b_data, c_data)):
            assert_allclose(new_tn[str(i)].data, arr.conj())

        # make sure original network unchanged
        for i, arr in enumerate((a_data, b_data, c_data)):
            assert_allclose(tn[str(i)].data, arr)

    def test_conj_inplace(self):
        a_data = np.random.randn(2, 3, 4) + 1.0j * np.random.randn(2, 3, 4)
        b_data = np.random.randn(3, 4, 5) + 1.0j * np.random.randn(3, 4, 5)
        c_data = np.random.randn(5, 2, 6) + 1.0j * np.random.randn(5, 2, 6)

        a = Tensor(a_data, inds=[0, 1, 2], tags={'red', 'I0'})
        b = Tensor(b_data, inds=[1, 2, 3], tags={'blue', 'I1'})
        c = Tensor(c_data, inds=[3, 0, 4], tags={'blue', 'I2'})

        tn = a & b & c
        tn.conj_()

        for i, arr in enumerate((a_data, b_data, c_data)):
            assert_allclose(tn[f"I{i}"].data, arr.conj())

    def test_multiply(self):
        a = rand_tensor((2, 3, 4), inds=['0', '1', '2'], tags='red')
        b = rand_tensor((3, 4, 5), inds=['1', '2', '3'], tags='blue')
        c = rand_tensor((5, 2, 6), inds=['3', '0', '4'], tags='blue')
        tn = a & b & c
        x1 = (tn & tn.H) ^ ...
        x2 = ((2 * tn) & tn.H) ^ ...
        assert_allclose(2 * x1, x2)

    def test_multiply_inplace(self):
        a = rand_tensor((2, 3, 4), inds=['0', '1', '2'], tags='red')
        b = rand_tensor((3, 4, 5), inds=['1', '2', '3'], tags='blue')
        c = rand_tensor((5, 2, 6), inds=['3', '0', '4'], tags='blue')
        tn = a & b & c
        x1 = (tn & tn.H) ^ ...
        tn *= 2
        x2 = (tn & tn.H) ^ ...
        assert_allclose(4 * x1, x2)

    def test_multiply_each(self):
        a = rand_tensor((2, 3, 4), inds=['0', '1', '2'], tags='red')
        b = rand_tensor((3, 4, 5), inds=['1', '2', '3'], tags='blue')
        c = rand_tensor((5, 2, 6), inds=['3', '0', '4'], tags='blue')
        tn = a & b & c
        x1 = (tn & tn.H) ^ ...
        x2 = (tn.multiply_each(2) & tn.H) ^ ...
        assert_allclose(2**3 * x1, x2)

    def test_divide(self):
        a = rand_tensor((2, 3, 4), inds=['0', '1', '2'], tags='red')
        b = rand_tensor((3, 4, 5), inds=['1', '2', '3'], tags='blue')
        c = rand_tensor((5, 2, 6), inds=['3', '0', '4'], tags='blue')
        tn = a & b & c
        x1 = (tn & tn.H) ^ ...
        x2 = ((tn / 2) & tn.H) ^ ...
        assert_allclose(x1 / 2, x2)

    def test_divide_inplace(self):
        a = rand_tensor((2, 3, 4), inds=['0', '1', '2'], tags='red')
        b = rand_tensor((3, 4, 5), inds=['1', '2', '3'], tags='blue')
        c = rand_tensor((5, 2, 6), inds=['3', '0', '4'], tags='blue')
        tn = a & b & c
        x1 = (tn & tn.H) ^ ...
        tn /= 2
        x2 = (tn & tn.H) ^ ...
        assert_allclose(x1 / 4, x2)

    def test_multiply_spread(self):
        a = rand_tensor([2, 2], inds=['a', 'b'], tags='A')
        b = Tensor(a.data, ['b', 'c'], tags='B')
        c = Tensor(a.data, ['c', 'd'], tags='C')
        tn = (a | b | c)
        tn.multiply_(-8j + 1 / 3, spread_over=3)
        assert_allclose(tn['A'].data, tn['B'].data)
        assert_allclose(tn['B'].data, tn['C'].data)

    def test_multiply_spread_neg_stays_real(self):
        a = rand_tensor([2, 2], inds=['a', 'b'], tags='A', dtype='float32')
        b = Tensor(a.data, ['b', 'c'], tags='B')
        c = Tensor(a.data, ['c', 'd'], tags='C')
        tn = (a | b | c)
        tn.multiply_(-1000)
        assert a.dtype == b.dtype == c.dtype == 'float32'
        assert_allclose(abs(tn['A'].data), abs(tn['B'].data))
        assert_allclose(abs(tn['B'].data), abs(tn['C'].data))

    def test_tensor_network_sum(self):
        A = qtn.TN_rand_reg(n=6, reg=3, D=2, phys_dim=2, dtype='complex')
        B = A.copy()
        B.randomize_()
        d1 = A.distance(B)
        AmB = qtn.tensor_network_sum(A, -1 * B)
        d2 = (AmB | AmB.H).contract(all)**0.5
        assert d1 == pytest.approx(d2)

    def test_contracting_tensors(self):
        a = rand_tensor((2, 3, 4), inds=[0, 1, 2], tags='red')
        b = rand_tensor((3, 4, 5), inds=[1, 2, 3], tags='blue')
        c = rand_tensor((5, 2, 6), inds=[3, 0, 4], tags='blue')

        a_b_c = a & b & c
        print(a_b_c)
        repr(a_b_c)

        assert isinstance(a_b_c, TensorNetwork)
        a_bc = a_b_c ^ 'blue'
        assert isinstance(a_bc, TensorNetwork)
        assert len(a_bc.tensors) == 2
        abc = a_bc ^ ['red', 'blue']
        assert isinstance(abc, Tensor)
        assert_allclose(abc.data, a_b_c.contract().data)

        assert len(a_b_c.tensors) == 3
        a_b_c ^= 'blue'
        assert len(a_b_c.tensors) == 2

    def test_cumulative_contract(self):
        a = rand_tensor((2, 3, 4), inds=[0, 1, 2], tags='red')
        b = rand_tensor((3, 4, 5), inds=[1, 2, 3], tags='blue')
        c = rand_tensor((5, 2, 6), inds=[3, 0, 4], tags='green')

        d = (a & b & c)
        d2 = d.copy()

        cd = d >> ['red', 'green', 'blue']
        assert cd.shape == (6,)
        assert cd.inds == (4,)

        # make sure inplace operations didn't effect original tensor
        for tag, names in d2.tag_map.items():
            assert d.tag_map[tag] == names

        # test inplace
        d >>= ['red', 'green', 'blue']
        assert isinstance(d, Tensor)

    def test_contract_with_slices(self):
        a = rand_tensor((2, 3, 4), inds=[0, 1, 2], tags='I0')
        b = rand_tensor((3, 4, 5), inds=[1, 2, 3], tags='I1')
        c = rand_tensor((5, 2, 6), inds=[3, 0, 4], tags='I2')
        d = rand_tensor((5, 2, 6), inds=[5, 6, 4], tags='I3')
        tn = TensorNetwork((a, b, c, d))
        tn.view_as_(TensorNetwork1D, L=4, site_tag_id='I{}')

        assert len((tn ^ slice(2)).tensors) == 3
        assert len((tn ^ slice(..., 1, -1)).tensors) == 3
        assert len((tn ^ slice(-1, 1)).tensors) == 3
        assert len((tn ^ slice(None, -2, -1)).tensors) == 3
        assert len((tn ^ slice(-2, 0)).tensors) == 3

    def test_contraction_info(self):
        a = qtn.rand_tensor((8, 8), ('a', 'b'))
        b = qtn.rand_tensor((8, 8), ('b', 'c'))
        c = qtn.rand_tensor((8, 8), ('c', 'd'))
        tn = a | b | c
        assert tn.contraction_width() == 6
        assert tn.contraction_cost() == 2 * 8**3

    @pytest.mark.parametrize('method', ('auto', 'dense', 'overlap'))
    def test_tensor_network_distance(self, method):
        n = 6
        A = qtn.TN_rand_reg(n=n, reg=3, D=2, phys_dim=2, dtype=complex)
        Ad = A.to_dense([f'k{i}' for i in range(n)])
        B = qtn.TN_rand_reg(n=6, reg=3, D=2, phys_dim=2, dtype=complex)
        Bd = B.to_dense([f'k{i}' for i in range(n)])
        d1 = np.linalg.norm(Ad - Bd)
        d2 = A.distance(B, method=method)
        assert d1 == pytest.approx(d2)

    @pytest.mark.parametrize('method,opts', (
        ('als', (('enforce_pos', False), ('solver', 'lstsq'))),
        ('als', (('enforce_pos', True),)),
        pytest.param('autodiff', (('distance_method', 'dense'),),
                     marks=autograd_mark),
        pytest.param('autodiff', (('distance_method', 'overlap'),),
                     marks=autograd_mark),
    ))
    def test_fit_mps(self, method, opts):
        k1 = qtn.MPS_rand_state(5, 3, seed=666)
        k2 = qtn.MPS_rand_state(5, 3, seed=667)
        assert k1.distance(k2) > 1e-3
        k1.fit_(k2, method=method, progbar=True, **dict(opts))
        assert k1.distance(k2) < 1e-3

    @pytest.mark.parametrize('method,opts', (
        ('als', (('enforce_pos', False),)),
        ('als', (('enforce_pos', True),)),
        pytest.param('autodiff', (('distance_method', 'dense'),),
                     marks=autograd_mark),
        pytest.param('autodiff', (('distance_method', 'overlap'),),
                     marks=autograd_mark),
    ))
    def test_fit_rand_reg(self, method, opts):
        r1 = qtn.TN_rand_reg(5, 4, D=2, seed=666, phys_dim=2)
        k2 = qtn.MPS_rand_state(5, 3, seed=667)
        assert r1.distance(k2) > 1e-3
        r1.fit_(k2, method=method, progbar=True, **dict(opts))
        assert r1.distance(k2) < 1e-3

    @pytest.mark.parametrize('method,opts', (
        ('als', (('enforce_pos', False),)),
        ('als', (('enforce_pos', True),)),
        pytest.param('autodiff', (('distance_method', 'dense'),),
                     marks=autograd_mark),
        pytest.param('autodiff', (('distance_method', 'overlap'),),
                     marks=autograd_mark),
    ))
    def test_fit_partial_tags(self, method, opts):
        k1 = qtn.MPS_rand_state(5, 3, seed=666)
        k2 = qtn.MPS_rand_state(5, 3, seed=667)
        d0 = k1.distance(k2)
        tags = ["I0", "I2", "I4"]
        k1f = k1.fit(k2, tol=1e-3, tags=tags,
                     method=method, progbar=True, **dict(opts))
        assert k1f.distance(k2) < d0
        assert (k1f[0] - k1[0]).norm() > 1e-12
        assert (k1f[1] - k1[1]).norm() < 1e-12
        assert (k1f[2] - k1[2]).norm() > 1e-12
        assert (k1f[3] - k1[3]).norm() < 1e-12
        assert (k1f[4] - k1[4]).norm() > 1e-12

    def test_reindex(self):
        a = Tensor(np.random.randn(2, 3, 4), inds=[0, 1, 2], tags='red')
        b = Tensor(np.random.randn(3, 4, 5), inds=[1, 2, 3], tags='blue')
        c = Tensor(np.random.randn(5, 2, 6), inds=[3, 0, 4], tags='green')

        a_b_c = (a & b & c)

        d = a_b_c.reindex({4: 'foo', 2: 'bar'})

        assert a_b_c.outer_inds() == (4,)
        assert d.outer_inds() == ('foo',)
        assert set(a_b_c.inner_inds()) == {0, 1, 2, 3}
        assert set(d.inner_inds()) == {0, 1, 'bar', 3}
        assert d.tensors[0].inds == (0, 1, 'bar')

        d = a_b_c.reindex_({4: 'foo', 2: 'bar'})

        assert a_b_c.outer_inds() == ('foo',)
        assert set(d.inner_inds()) == {0, 1, 'bar', 3}
        assert d.tensors[0].inds == (0, 1, 'bar')

    def test_add_tag(self):
        a = rand_tensor((2, 3, 4), inds='abc', tags={'red'})
        b = rand_tensor((2, 3, 4), inds='abc', tags={'blue'})
        tn = a & b
        tn.add_tag('green')
        assert 'green' in tn.tag_map
        assert 'green' in tn['red'].tags
        assert 'green' in tn['blue'].tags
        tn.add_tag('blue')
        for t in tn:
            assert 'blue' in t.tags

    def test_index_by_site(self):
        a_data = np.random.randn(2, 3, 4)
        b_data = np.random.randn(2, 3, 4)
        a = Tensor(a_data, inds='abc', tags={'I0'})
        b = Tensor(b_data, inds='abc', tags={'I1'})
        tn = TensorNetwork((a, b))
        tn.view_as_(TensorNetwork1D, L=2, site_tag_id='I{}')
        assert_allclose(tn[0].data, a_data)
        new_data = np.random.randn(2, 3, 4)
        tn[1] = Tensor(new_data, inds='abc', tags={'I1', 'red'})
        assert_allclose(tn['I1'].data, new_data)
        assert 'red' in tn['I1'].tags

    def test_set_data_in_tensor(self):
        a_data = np.random.randn(2, 3, 4)
        b_data = np.random.randn(2, 3, 4)
        a = Tensor(a_data, inds='abc', tags={'I0'})
        b = Tensor(b_data, inds='abc', tags={'I1'})
        tn = TensorNetwork((a, b))
        tn.view_as_(TensorNetwork1D, L=2, site_tag_id='I{}')
        assert_allclose(tn[0].data, a_data)
        new_data = np.random.randn(2, 3, 4)
        tn[1].modify(data=new_data)
        assert_allclose(tn['I1'].data, new_data)

    def test_make_tids_consecutive_combining_with_no_check_collisions(self):
        p1 = MPS_rand_state(5, 3, phys_dim=3)
        p2 = MPS_rand_state(5, 3, phys_dim=3)
        p2.make_tids_consecutive(tid0=5)
        # shouldn't need to check any collisions
        tn = TensorNetwork((p1, p2), check_collisions=False)
        # test can contract
        assert 0 < abs(tn ^ ...) < 1

    def test_retagging(self):
        x = rand_tensor((2, 4), inds='ab', tags={'X', 'I0'})
        y = rand_tensor((4, 2, 5), inds='bcd', tags={'Y', 'I1'})
        z = rand_tensor((5, 3), inds='de', tags={'Z', 'I2'})
        tn = TensorNetwork((x, y, z))
        tn.retag_({"I0": "I1", "I1": "I2", "I2": "I3", "Z": "A"})
        assert set(tn.tag_map.keys()) == {'X', 'I1', 'I2', 'I3', 'Y', 'A'}

    def test_squeeze(self):
        A, B, C = (rand_tensor((1, 2, 3), 'abc', tags=['I0']),
                   rand_tensor((2, 3, 4), 'bcd', tags=['I1']),
                   rand_tensor((4, 1, 1), 'dae', tags=['I2']))
        tn = A & B & C

        x1 = tn ^ ...
        stn = tn.squeeze()

        assert tn['I0'].shape == (1, 2, 3)
        assert tn['I1'].shape == (2, 3, 4)
        assert tn['I2'].shape == (4, 1, 1)

        assert stn['I0'].shape == (2, 3)
        assert stn['I1'].shape == (2, 3, 4)
        assert stn['I2'].shape == (4,)

        x2 = stn ^ ...
        assert_allclose(x1.data, x2)  # x2 should be scalar already

    def test_tensors_sorted(self):
        tn1, tn2 = TensorNetwork([]), TensorNetwork([])
        A, B, C = (rand_tensor((1, 2, 3), 'abc', tags=['I0']),
                   rand_tensor((2, 3, 4), 'bcd', tags=['I1']),
                   rand_tensor((4, 1, 1), 'dae', tags=['I2']))

        tn1 &= A
        tn1 &= B
        tn1 &= C

        tn2 &= C
        tn2 &= A
        tn2 &= B

        for t1, t2 in zip(tn1.tensors_sorted(), tn2.tensors_sorted()):
            assert t1.tags == t2.tags
            assert t1.almost_equals(t2)

    def test_select_tensors_mode(self):
        A, B, C = (rand_tensor((2, 2), 'ab', tags={'0', 'X'}),
                   rand_tensor((2, 2), 'bc', tags={'1', 'X', 'Y'}),
                   rand_tensor((2, 3), 'cd', tags={'2', 'Y'}))
        tn = A & B & C

        ts = tn.select_tensors(('X', 'Y'), which='all')
        assert len(ts) == 1
        assert not any(map(A.almost_equals, ts))
        assert any(map(B.almost_equals, ts))
        assert not any(map(C.almost_equals, ts))

        ts = tn.select_tensors(('X', 'Y'), which='any')
        assert len(ts) == 3
        assert any(map(A.almost_equals, ts))
        assert any(map(B.almost_equals, ts))
        assert any(map(C.almost_equals, ts))

    def test_replace_with_identity(self):
        A, B, C, D = (rand_tensor((2, 3, 4), 'abc', tags=['I0']),
                      rand_tensor((4, 5, 6), 'cde', tags=['I1']),
                      rand_tensor((5, 6, 7), 'def', tags=['I2']),
                      rand_tensor((7,), 'f', tags=['I3']))

        tn = (A & B & C & D)

        with pytest.raises(ValueError):
            tn.replace_with_identity(('I1', 'I2'), inplace=True)

        tn['I2'] = rand_tensor((5, 6, 4), 'def', tags=['I2'])
        tn['I3'] = rand_tensor((4,), 'f', tags=['I3'])

        tn1 = tn.replace_with_identity(('I1', 'I2'))
        assert len(tn1.tensors) == 2
        x = tn1 ^ ...
        assert set(x.inds) == {'a', 'b'}

        A, B, C = (rand_tensor((2, 2), 'ab', tags={'0'}),
                   rand_tensor((2, 2), 'bc', tags={'1'}),
                   rand_tensor((2, 3), 'cd', tags={'2'}))

        tn = A & B & C

        tn2 = tn.replace_with_identity('1')
        assert len(tn2.tensors) == 2
        x = tn2 ^ ...
        assert set(x.inds) == {'a', 'd'}

    def test_partition(self):
        k = MPS_rand_state(10, 7, site_tag_id='Q{}')
        where = [f'Q{i}' for i in range(10) if i % 2 == 1]
        k.add_tag('odd', where=where, which='any')

        tn_even, tn_odd = k.partition('odd')

        assert len(tn_even.tensors) == len(tn_odd.tensors) == 5

        assert tn_even.site_tag_id == 'Q{}'
        assert tn_odd.site_tag_id == 'Q{}'

        assert (tn_even & tn_odd).sites == tuple(range(10))

    def test_subgraphs(_):
        k1 = MPS_rand_state(6, 7, site_ind_id='a{}')
        k2 = MPS_rand_state(8, 7, site_ind_id='b{}')
        tn = k1 | k2
        s1, s2 = tn.subgraphs()
        assert {s1.num_tensors, s2.num_tensors} == {6, 8}

    def test_compress_multibond(self):
        A = rand_tensor((7, 2, 2), 'abc', tags='A')
        A.expand_ind('c', 3)
        B = rand_tensor((3, 2, 7), 'cbd', tags='B')
        x0 = (A & B).trace('a', 'd')
        qtn.tensor_compress_bond(A, B, absorb='left')
        A.transpose_('a', 'b')
        assert A.shape == (7, 4)
        B.transpose_('b', 'd')
        assert B.shape == (4, 7)
        assert B.H @ B == pytest.approx(4)
        x1 = (A & B).trace('a', 'd')
        assert x1 == pytest.approx(x0)

    def test_canonize_multibond(self):
        A = rand_tensor((3, 4, 5), 'abc', tags='A')
        assert A.H @ A != pytest.approx(3)
        B = rand_tensor((5, 4, 3), 'cbd', tags='B')
        x0 = (A & B).trace('a', 'd')
        qtn.tensor_canonize_bond(A, B)
        assert A.shape == (3, 3)
        assert B.shape == (3, 3)
        assert A.H @ A == pytest.approx(3)
        x1 = (A & B).trace('a', 'd')
        assert x1 == pytest.approx(x0)

    @pytest.mark.parametrize("method", ['svd', 'eig', 'isvd', 'svds', 'rsvd'])
    def test_compress_between(self, method):
        A = rand_tensor((3, 4, 5), 'abd', tags={'T1'})
        A.expand_ind('d', 10)
        B = rand_tensor((5, 6), 'dc', tags={'T2'})
        B.expand_ind('d', 10)
        tn = A | B
        assert A.shared_bond_size(B) == 10
        tn.compress_between('T1', 'T2', method=method)
        assert A.shared_bond_size(B) == 5

    @pytest.mark.parametrize("method", ['svd', 'eig', 'isvd', 'svds', 'rsvd'])
    def test_compress_all(self, method):
        k = MPS_rand_state(10, 7)
        k += k
        k /= 2
        k.compress_all_(max_bond=7, method=method)
        assert k.max_bond() == 7
        assert_allclose(k.H @ k, 1.0)

    def test_canonize_between(self):
        k = MPS_rand_state(4, 3)
        k.canonize_between('I1', 'I2')
        assert k.H @ k == pytest.approx(1)
        t = k[1]
        assert t.H @ t == pytest.approx(3)
        t = k[2]
        assert t.H @ t != pytest.approx(3)
        k.canonize_between('I2', 'I1')
        assert k.H @ k == pytest.approx(1)
        t = k[1]
        assert t.H @ t != pytest.approx(3)
        t = k[2]
        assert t.H @ t == pytest.approx(3)

    def test_canonize_around(self):
        # make a small tree tensor network
        #
        #             U2--                         v--
        #             |                            |
        #             U1--         ==>             v--        etc
        #             |                            |
        #   L2---L1---C---R1---R2        >---->---->----O---<
        #  /     |    |    |    \       /     |    |    |    \
        #
        C = qtn.rand_tensor([2], inds=['kC'], tags='C', dtype=complex)

        # left arm
        L1 = qtn.rand_tensor([2], inds=['kL1'], tags='L1', dtype=complex)
        qtn.new_bond(C, L1, size=7)
        L2 = qtn.rand_tensor([2], inds=['kL2'], tags='L2', dtype=complex)
        qtn.new_bond(L1, L2, size=7)

        # right arm
        R1 = qtn.rand_tensor([2], inds=['kR1'], tags='R1', dtype=complex)
        qtn.new_bond(C, R1, size=7)
        R2 = qtn.rand_tensor([2], inds=['kR2'], tags='R2', dtype=complex)
        qtn.new_bond(R1, R2, size=7)

        # upper arm
        U1 = qtn.rand_tensor([2], inds=['kU1'], tags='U1', dtype=complex)
        qtn.new_bond(C, U1, size=7)
        U2 = qtn.rand_tensor([2], inds=['kU2'], tags='U2', dtype=complex)
        qtn.new_bond(U1, U2, size=7)

        # make the TN and randomize the data then normalize
        ttn = qtn.TensorNetwork([C, L1, L2, R1, R2, U1, U2])
        ttn.randomize_()
        ttn /= (ttn.H @ ttn)**0.5
        assert ttn.H @ ttn == pytest.approx(1.0)

        # test max distance
        ttn.canonize_around_('C', max_distance=1)
        assert ttn.H @ ttn == pytest.approx(1.0)
        assert ttn['C'].H @ ttn['C'] != pytest.approx(1.0)

        # tensors one-away from center should be isometries
        for tg in ['L1', 'R1', 'U1']:
            assert ttn[tg].H @ ttn[tg] == pytest.approx(7)
        # tensors two-away from center should be random
        for tg in ['L2', 'R2', 'U2']:
            assert ttn[tg].H @ ttn[tg] != pytest.approx(2)

        ttn.canonize_around_('C', max_distance=2)
        for tg in ['L2', 'R2', 'U2']:
            assert ttn[tg].H @ ttn[tg] == pytest.approx(2)

        # test can set the orthogonality center anywhere
        for tg in ['C', 'L1', 'L2', 'R1', 'R2', 'U1', 'U2']:
            ttn.canonize_around_(tg)
            assert ttn.H @ ttn == pytest.approx(1.0)
            assert ttn[tg].H @ ttn[tg] == pytest.approx(1.0)

            # tensors two-away from center should now be isometries
            for far_tg in ['L2', 'R2', 'U2']:
                if far_tg != tg:
                    ttn[far_tg].H @ ttn[far_tg] == pytest.approx(2)

    def test_tn_split_tensor(self):
        mps = MPS_rand_state(4, 3)
        right_inds = bonds(mps[1], mps[2])
        mps.split_tensor(1, left_inds=None, right_inds=right_inds, rtags='X')
        assert mps.num_tensors == 5
        assert mps['X'].shape == (3, 3)
        assert mps.H @ mps == pytest.approx(1.0)

    def test_insert_operator(self):
        p = MPS_rand_state(3, 7, tags='KET')
        q = p.H.retag({'KET': 'BRA'})
        qp = q & p
        sz = qu.spin_operator('z').real
        qp.insert_operator(sz, ('KET', 'I1'), ('BRA', 'I1'),
                           tags='SZ', inplace=True)
        assert 'SZ' in qp.tags
        assert len(qp.tensors) == 7
        x1 = qp ^ all
        x2 = qu.expec(p.to_dense(), qu.ikron(sz, [2, 2, 2], inds=1))
        assert x1 == pytest.approx(x2)

    @pytest.mark.parametrize("dtype", (float, complex))
    def test_insert_gauge(self, dtype):
        k = MPS_rand_state(10, 7, dtype=dtype, normalize=False)
        kU = k.copy()

        U = rand_tensor((7, 7), dtype=dtype, inds='ab').data
        kU.insert_gauge(U, 4, 5)

        assert k[3].almost_equals(kU[3])
        assert not k[4].almost_equals(kU[4])
        assert not k[5].almost_equals(kU[5])
        assert k[6].almost_equals(kU[6])

        assert k[4].inds == kU[4].inds
        assert k[5].inds == kU[5].inds

        assert_allclose(k.H @ k, kU.H @ kU)

    def test_fuse_multibonds(self):
        x = rand_tensor((2, 2, 2), ['a', 'b', 'c'])
        y = rand_tensor((2, 2, 2, 2), ['b', 'c', 'd', 'e'])
        z = rand_tensor((2, 2, 2), ['a', 'e', 'd'])
        tn = (x & y & z)
        assert len(tn.inner_inds()) == 5
        tn.fuse_multibonds(inplace=True)
        assert len(tn.inner_inds()) == 3

    def test_draw(self):
        import matplotlib
        from matplotlib import pyplot as plt
        matplotlib.use('Template')
        k = MPS_rand_state(10, 7, normalize=False)
        fig = k.draw(color=['I0', 'I2'], return_fig=True)
        plt.close(fig)

    def test_draw_with_fixed_pos(self):
        import matplotlib
        from matplotlib import pyplot as plt
        matplotlib.use('Template')
        n = 7
        p = MPS_rand_state(n, 7, tags='KET')
        q = MPS_rand_state(n, 7, tags='BRA')
        fix = {**{('KET', f'I{i}'): (i, 0) for i in range(n)},
               **{('BRA', f'I{i}'): (i, 1) for i in range(n)}}
        fig = (q | p).draw(color=['KET', 'BRA'], fix=fix, return_fig=True)
        plt.close(fig)

    def test_pickle(self):
        import tempfile
        import os

        pytest.importorskip("joblib")

        tn = MPS_rand_state(10, 7, tags='KET')

        with tempfile.TemporaryDirectory() as tdir:
            fname = os.path.join(tdir, "tn.dmp")
            qu.save_to_disk(tn, fname)
            tn2 = qu.load_from_disk(fname)

        assert tn.H @ tn2 == pytest.approx(1.0)

        assert all(hash(tn) not in t.owners for t in tn2)
        assert all(hash(tn2) in t.owners for t in tn2)

    @pytest.mark.parametrize('dtype', [None, 'float32', 'complex128'])
    def test_randomize(self, dtype):
        psi = MPS_rand_state(5, 3, dtype='float64')
        x1 = psi.H @ psi
        psi.randomize_(seed=42, dtype=dtype)
        x2 = psi.H @ psi
        assert x1 != pytest.approx(x2)
        if dtype is None:
            assert psi.dtype == 'float64'
        else:
            assert psi.dtype == dtype
        psi.randomize_(seed=42, dtype=dtype)
        x3 = psi.H @ psi
        assert x2 == pytest.approx(x3)

    @pytest.mark.parametrize('dtype', ['float32', 'complex128'])
    @pytest.mark.parametrize('value', [None, 42])
    def test_equalize_norms(self, dtype, value):
        psi = MPS_rand_state(5, 3, dtype=dtype)
        psi.randomize_(seed=42)
        x_exp = psi.H @ psi
        norms = [t.norm() for t in psi]
        psi.equalize_norms_(value)
        enorms = [t.norm() for t in psi]
        if value is None:
            assert all(n1 != n2 for n1, n2 in zip(norms, enorms))
            assert psi.H @ psi == pytest.approx(x_exp)
        else:
            assert all(n1 == pytest.approx(value) for n1 in enorms)
            assert (
                (psi.H @ psi) * 10**(2 * psi.exponent) == pytest.approx(x_exp)
            )

    @pytest.mark.parametrize("append", [None, '*'])
    def test_mangle_inner(self, append):
        a = MPS_rand_state(6, 3)
        b = a.copy()
        assert tuple(a.ind_map) == tuple(b.ind_map)
        b.mangle_inner_(append)
        assert tuple(a.ind_map) != tuple(b.ind_map)
        ab = a & b
        assert all(ix in ab.ind_map for ix in a.ind_map)
        assert all(ix in ab.ind_map for ix in b.ind_map)

    @pytest.mark.parametrize("mode", ["manual", "dense", "mps", "tree"])
    def test_hyperind_resolve(self, mode):
        import networkx as nx
        import random
        import collections

        # create a random interaction ising model
        G = nx.watts_strogatz_graph(10, 4, 0.5, seed=666)
        edges = tuple(G.edges)
        js = collections.defaultdict(random.random)
        htn = qtn.HTN_classical_partition_function_from_edges(
            edges, j=lambda i, j: js[frozenset((i, j))], beta=0.22, h=0.04)
        Zh = htn.contract(all, output_inds=())
        assert len(htn.get_hyperinds(output_inds=())) == 10
        if mode == "manual":
            # resolve manually
            tn = qtn.TN_classical_partition_function_from_edges(
                edges, j=lambda i, j: js[frozenset((i, j))], beta=0.22, h=0.04)
        else:
            tn = htn.hyperinds_resolve(mode)
        assert len(tn.get_hyperinds()) == 0
        Z = tn.contract(all, output_inds=())
        assert Z == pytest.approx(Zh)
        assert max(map(len, tn.ind_map.values())) == 2


class TestTensorNetworkSimplifications:

    def test_rank_simplify(self):
        A = rand_tensor([2, 2, 3], 'abc', tags='A')
        B = rand_tensor([3, 2], 'cd', tags='B')
        C = rand_tensor([2, 2, 2], 'def', tags='C')
        tn = A & B & C
        tn_s = tn.rank_simplify()
        assert tn.num_tensors == 3
        assert tn_s.num_tensors == 2
        assert (tn ^ all).almost_equals(tn_s ^ all)
        # checl that 'B' was absorbed into 'A' not 'C'
        assert set(tn_s['B'].tags) == {'A', 'B'}

    def test_rank_simplify_single_ind(self):
        ts = [rand_tensor([2], 'a') for _ in range(100)]
        tn = TensorNetwork(ts)
        assert len(tn.ind_map) == 1
        assert len(tn.tensor_map) == 100
        tn.rank_simplify_()
        assert len(tn.tensor_map) == 1

    def test_diagonal_reduce(self):
        A = rand_tensor([2, 2], 'ab', dtype=complex)
        B = Tensor([[3j, 0.], [0., 4j]], 'bc')
        C = rand_tensor([2, 2], 'ca', dtype=complex)
        tn = A & B & C
        tn_s = tn.diagonal_reduce()
        assert tn.num_indices == 3
        assert tn_s.num_indices == 2
        assert tn ^ all == pytest.approx(tn_s.contract(all, output_inds=[]))

    def test_antidiag_gauge(self):
        A = rand_tensor([2, 2], 'ab', dtype=complex)
        B = Tensor([[0., 3j], [4j, 0.]], 'bc')
        C = rand_tensor([2, 2], 'ca', dtype=complex)
        tn = A & B & C
        assert tn.num_indices == 3
        # can't use diagonal reduction yet
        assert tn.diagonal_reduce().num_indices == 3
        # initial gauge doesn't change indices
        tn_a = tn.antidiag_gauge()
        assert tn_a.num_indices == 3
        # but allows the diagonal reduction
        tn_ad = tn_a.diagonal_reduce()
        assert tn_ad.num_indices == 2
        assert tn ^ all == pytest.approx(tn_ad.contract(all, output_inds=[]))

    def test_column_reduce(self):
        A = rand_tensor([2, 3], 'ab')
        A.new_ind('c', size=4, axis=-2)
        B = rand_tensor([4, 5, 6], 'cde')
        tn = A & B
        assert tn.num_indices == 5
        tn_s = tn.column_reduce()
        assert tn_s.num_indices == 4
        assert (tn ^ all).almost_equals(tn_s ^ all)


class TestTensorNetworkAsLinearOperator:

    @pytest.mark.parametrize('optimize', ['auto', 'auto-hq'])
    def test_against_dense(self, optimize):
        A, B, C, D = (
            rand_tensor([3, 5, 5], 'aef'),
            rand_tensor([3, 5, 5], 'beg'),
            rand_tensor([3, 5, 5], 'cfh'),
            rand_tensor([3, 5, 5], 'dhg'),
        )

        tn = A & B & C & D
        tn_lo = tn.aslinearoperator(('a', 'b'), ('c', 'd'), optimize=optimize)
        tn_d = tn.to_dense(['a', 'b'], ['c', 'd'])

        u, s, v = qu.svds(tn_lo, k=5, backend='scipy')
        ud, sd, vd = qu.svds(tn_d, k=5, backend='scipy')

        assert_allclose(s, sd)

        # test matmat
        X = np.random.randn(9, 8) + 1.0j * np.random.randn(9, 8)
        assert_allclose(tn_lo.dot(X), tn_d.dot(X))

    def test_trace_array_function_interface(self):
        tn = qtn.TensorNetwork((
            rand_tensor([3, 5, 5], 'aef'),
            rand_tensor([3, 5, 5], 'beg'),
            rand_tensor([3, 5, 5], 'cfh'),
            rand_tensor([3, 5, 5], 'dhg'),
        ))
        tn_lo = tn.aslinearoperator(('a', 'b'), ('c', 'd'))
        tn_d = tn.to_dense(['a', 'b'], ['c', 'd'])
        assert np.trace(tn_lo) == pytest.approx(np.trace(tn_d))

    @pytest.mark.parametrize("dtype", (float, complex))
    @pytest.mark.parametrize("method", ('isvd', 'rsvd'))
    def test_replace_with_svd_using_linear_operator(self, dtype, method):
        k = MPS_rand_state(100, 10, dtype=dtype, cyclic=True)
        b = k.H
        b.expand_bond_dimension(11)
        k.add_tag('_KET')
        b.add_tag('_BRA')
        tn = b & k

        x1 = tn ^ ...

        ul, = tn['_KET', 'I1'].bonds(tn['_KET', 'I2'])
        ll, = tn['_BRA', 'I1'].bonds(tn['_BRA', 'I2'])

        where = [f'I{i}' for i in range(2, 40)]

        tn.replace_with_svd(where, left_inds=(ul, ll), eps=1e-3, method=method,
                            inplace=True, ltags='_U', rtags='_V')

        x2 = tn ^ ...

        # check ltags and rtags have gone in
        assert isinstance(tn['_U'], Tensor)
        assert isinstance(tn['_V'], Tensor)

        assert_allclose(x1, x2, rtol=1e-4)

    def test_TNLinearOperator1D(self):
        p = MPS_rand_state(40, 10, dtype=complex)
        pp = p.H & p
        start, stop = 10, 30
        lix = bonds(pp[start - 1], pp[start])
        rix = bonds(pp[stop - 1], pp[stop])

        sec = pp[start:stop]

        A = TNLinearOperator1D(sec, lix, rix, start, stop)
        B = sec.aslinearoperator(lix, rix)

        s1 = spla.svds(A)[1]
        s2 = spla.svds(B)[1]

        assert_allclose(s1, s2)



"""Tools for constructing MERA for arbitrary geometry.

TODO:

    - [ ] 2D, 3D MERA classes
    - [ ] general strategies for arbitrary geometries
    - [ ] layer_tag? and hanling of other attributes
    - [ ] handle dangling case
    - [ ] invariant generators?

 DONE:

    - [x] layer_gate methods for arbitrary geometry
    - [x] 1D: generic way to handle finite and open boundary conditions
    - [x] hook into other arbgeom infrastructure for computing rdms etc

"""
import itertools
from quimb.tensor.tensor_arbgeom import (
    TensorNetworkGenVector,
    functools,
    oset,
    tags_to_oset,
    rand_uuid,
    oset_union,
    IsoTensor,
    check_opt,
    prod,
    do,
    _compute_expecs_maybe_in_parallel,
    _tn_local_expectation,
)
from quimb.tensor.tensor_1d import TensorNetwork1DVector
from quimb.utils import partition


class TensorNetworkGenIso(TensorNetworkGenVector):
    """A class for building generic 'isometric' or MERA like tensor network
    states with arbitrary geometry. After supplying the underyling `sites` of
    the problem - which can be an arbitrary sequence of hashable objects - one
    places either unitaries, isometries or tree tensors layered above groups of
    sites. The isometric and tree tensors effectively coarse grain blocks into
    a single new site, and the unitaries generally 'disentangle' between
    blocks.
    """

    _EXTRA_PROPS = (
        '_site_tag_id',
        '_sites',
        '_site_ind_id',
        '_layer_ind_id',
    )

    @classmethod
    def empty(
        cls,
        sites,
        phys_dim=2,
        site_tag_id="I{}",
        site_ind_id="k{}",
        layer_ind_id="l{}",
    ):
        new = object.__new__(cls)
        new.phys_dim = phys_dim
        new._sites = tuple(sites)
        new._site_tag_id = site_tag_id
        new._site_ind_id = site_ind_id
        new._layer_ind_id = layer_ind_id

        new._open_upper_sites = oset(new._sites)
        new._open_lower_sites = oset(new._sites)

        super().__init__(new, ())
        return new

    @property
    def layer_ind_id(self):
        return self._layer_ind_id

    def layer_ind(self, site):
        return self._layer_ind_id.format(site)

    def layer_gate_raw(
        self,
        G,
        where,
        iso=True,
        new_sites=None,
        tags=None,
        all_site_tags=None,
    ):
        """Build out this MERA by placing either a new unitary, isometry or
        tree tensor, given by ``G``, at the sites given by ``where``. This
        handles propagating the lightcone of tags and marking the correct
        indices of the ``IsoTensor`` as ``left_inds``.

        Parameters
        ----------
        G : array_like
            The raw array to place at the sites. Its shape determines whether
            it is a unitary or isometry/tree. It should have ``k + len(where)``
            dimensions. For a unitary ``k == len(where)``. If it is an
            isometry/tree, ``k`` will generally be ``1``, or ``0`` to 'cap' the
            MERA. The rightmost indices are those attached to the current open
            layer indices.
        where : sequence of hashable
            The sites to layer the tensor above.
        iso : bool, optional
            Whether to declare the tensor as an unitary/isometry by marking
            the left indices. If ``iso = False`` (a 'tree' tensor) then one
            should have ``k <= 1``. Once you have such a 'tree' tensor you
            cannot place isometries or unitaries above it. It will also have
            the lightcone tags of every site. Technically one could place
            'PEPS' style tensor with ``iso = False`` and ``k > 1`` but some
            methods might break.
        new_sites : sequence of hashable, optional
            Which sites to make new open sites. If not given, defaults to the
            first ``k`` sites in ``where``.
        tags : sequence of str, optional
            Custom tags to add to the new tensor, in addition to the
            automatically generated site tags.
        all_site_tags : sequence of str, optional
            For performance, supply all site tags to avoid recomputing them.
        """
        if all_site_tags is None:
            all_site_tags = oset(map(self.site_tag, self.gen_site_coos()))

        # work out 'lower' tensor indices
        nbelow = len(where)
        below_ix = []
        reindex_map = {}
        tags = tags_to_oset(tags)
        for site in where:
            if site in self._open_lower_sites:
                # this is the first tensor placed above site
                below_ix.append(self.site_ind(site))
                self._open_lower_sites.remove(site)
                tags.add(self.site_tag(site))
            else:
                # tensor is being placed above existing tensor
                current_lower_ix = self.layer_ind(site)
                new_lower_ix = rand_uuid()
                reindex_map[current_lower_ix] = new_lower_ix
                below_ix.append(new_lower_ix)

        # work out 'upper' tensor indices
        nabove = len(G.shape) - nbelow
        if new_sites is None:
            new_sites = where[:nabove]
        above_ix = [self.layer_ind(site) for site in new_sites]

        for site in where[nabove:]:
            # if tensor is not a unitary then some upper sites are removed
            self._open_upper_sites.remove(site)

        # want to propagate just site tags from tensors below
        old_tags = oset_union(t.tags for t in self._inds_get(*reindex_map))

        if iso and 'TREE' in old_tags:
            raise ValueError(
                "You can't place isometric tensors above tree tensors."
            )

        if not iso:
            # tensor is in lightcone of all sites
            tags |= all_site_tags
            tags.add('TREE')
            left_inds = None
            if nabove > 1:
                import warnings
                warnings.warn(
                    "You are placing a tensor which is neither "
                    "isometric/unitary or a tree. Some methods might break."
                )
        else:
            # just want site tags present on tensors below
            tags |= (old_tags & all_site_tags)
            if nbelow == nabove:
                tags.add('UNI')
            else:
                tags.add('ISO')
            if nabove == 0:
                tags.add('CAP')
            left_inds = below_ix

        # rewire and add tensor
        self.reindex_(reindex_map)
        self |= IsoTensor(
            data=G,
            inds=below_ix + above_ix,
            left_inds=left_inds,
            tags=tags,
        )

    def layer_gate_fill_fn(
        self,
        fill_fn,
        operation,
        where,
        max_bond,
        new_sites=None,
        tags=None,
        all_site_tags=None,
    ):
        """Build out this MERA by placing either a new unitary, isometry or
        tree tensor at sites ``where``, generating the data array using
        ``fill_fn`` and maximum bond dimension ``max_bond``.

        Parameters
        ----------
        fill_fn : callable
            A function with signature ``fill_fn(shape) -> array_like``.
        operation : {"iso", "uni", "tree"}
            The type of tensor to place.
        where : sequence of hashable
            The sites to layer the tensor above.
        max_bond : int
            The maximum bond dimension of the tensor. This only applies for
            isometries and trees and when the product of the lower dimensions
            is greater than ``max_bond``.
        new_sites : sequence of hashable, optional
            Which sites to make new open sites. If not given, defaults to the
            first ``k`` sites in ``where``.
        tags : sequence of str, optional
            Custom tags to add to the new tensor, in addition to the
            automatically generated site tags.
        all_site_tags : sequence of str, optional
            For performance, supply all site tags to avoid recomputing them.

        See Also
        --------
        layer_gate_raw
        """
        check_opt("operation", operation, ("iso", "uni", "tree", "cap"))

        shape = []
        for site in where:
            if site in self._open_lower_sites:
                shape.append(self.phys_dim)
            else:
                shape.append(self.ind_size(self.layer_ind(site)))

        if operation == "uni":
            # unitary, map is shape preserving
            shape = (*shape, *shape)
        elif operation in ("iso", "tree"):
            current_size = prod(shape)
            new_size = min(current_size, max_bond)
            shape = (*shape, new_size)
        else:  # "cap"
            shape = tuple(shape)

        G = fill_fn(shape)
        self.layer_gate_raw(
            G,
            where,
            new_sites=new_sites,
            tags=tags,
            all_site_tags=all_site_tags,
            iso=operation != "tree",
        )

    def partial_trace(
        self,
        keep,
        optimize='auto-hq',
        rehearse=False,
        **contract_opts,
    ):
        """Partial trace out all sites except those in ``keep``, making use of
        the lightcone structure of the MERA.

        Parameters
        ----------
        keep : sequence of hashable
            The sites to keep.
        optimize : str or PathOptimzer, optional
            The contraction ordering strategy to use.
        rehearse : {False, "tn", "tree"}, optional
            Whether to rehearse the contraction rather than actually performing
            it. If:

            - ``False``: perform the contraction and return the reduced density
              matrix,
            - "tn": just the lightcone tensor network is returned,
            - "tree": just the contraction tree that will be used is returned.

        contract_opts
            Additional options to pass to
            :func:`~quimb.tensor.tensor_core.tensor_contract`.

        Returns
        -------
        array_like
            The reduced density matrix on sites ``keep``.
        """
        tags = tuple(map(self.site_tag, keep))
        k = self.select_any(tags, virtual=False)

        kix = tuple(map(self.site_ind, keep))
        bix = tuple(f'b{site}' for site in keep)
        b = k.reindex(dict(zip(kix, bix))).conj_()
        tn = b | k

        if rehearse == 'tn':
            return tn

        if rehearse == 'tree':
            return tn.contraction_tree(
                output_inds=bix + kix,
                optimize=optimize,
                **contract_opts
            )

        return tn.to_dense(
            bix, kix,
            optimize=optimize,
            **contract_opts,
        )

    def local_expectation(
        self,
        G,
        where,
        optimize='auto-hq',
        rehearse=False,
        **contract_opts,
    ):
        """Compute the expectation value of a local operator ``G`` at sites
        ``where``. This is done by contracting the lightcone tensor network
        to form the reduced density matrix, before taking the trace with
        ``G``.

        Parameters
        ----------
        G : array_like
            The local operator to compute the expectation value of.
        where : sequence of hashable
            The sites to compute the expectation value at.
        optimize : str or PathOptimzer, optional
            The contraction ordering strategy to use.
        rehearse : {False, "tn", "tree"}, optional
            Whether to rehearse the contraction rather than actually performing
            it. See :meth:`~quimb.tensor.mera.MERA.partial_trace` for details.
        contract_opts
            Additional options to pass to
            :func:`~quimb.tensor.tensor_core.tensor_contract`.

        Returns
        -------
        float
            The expectation value of ``G`` at sites ``where``.

        See Also
        --------
        partial_trace
        """
        rho = self.partial_trace(
            keep=where,
            optimize=optimize,
            rehearse=rehearse,
            **contract_opts,
        )

        if rehearse:
            return rho

        return do("tensordot", rho, G, axes=((0, 1), (1, 0)))

    def compute_local_expectation(
        self,
        terms,
        optimize='auto-hq',
        return_all=False,
        rehearse=False,
        executor=None,
        progbar=False,
        **contract_opts,
    ):
        """Compute the expectation value of a collection of local operators
        ``terms`` at sites ``where``. This is done by contracting the lightcone
        tensor network to form the reduced density matrices, before taking the
        trace with each ``G`` in ``terms``.

        Parameters
        ----------
        terms : dict[tuple[hashable], array_like]
            The local operators to compute the expectation value of, keyed by
            the sites they act on.
        optimize : str or PathOptimzer, optional
            The contraction ordering strategy to use.
        return_all : bool, optional
            Whether to return all the expectation values, or just the sum.
        rehearse : {False, "tn", "tree"}, optional
            Whether to rehearse the contraction rather than actually performing
            it. See :meth:`~quimb.tensor.mera.MERA.partial_trace` for details.
        executor : Executor, optional
            The executor to use for parallelism.
        progbar : bool, optional
            Whether to show a progress bar.
        contract_opts
            Additional options to pass to
            :func:`~quimb.tensor.tensor_core.tensor_contract`.
        """
        return _compute_expecs_maybe_in_parallel(
            fn=_tn_local_expectation,
            tn=self,
            terms=terms,
            return_all=return_all,
            executor=executor,
            progbar=progbar,
            optimize=optimize,
            rehearse=rehearse,
            **contract_opts,
        )

    def expand_bond_dimension(
        self,
        new_bond_dim,
        rand_strength=0.0,
        inds_to_expand=None,
        inplace=False,
    ):
        """Expand the maxmimum bond dimension of this isometric tensor network
        to ``new_bond_dim``. Unlike
        :meth:`~quimb.tensor.tensor_core.TensorNetwork.expand_bond_dimension`
        this proceeds from the physical indices upwards, and only increases a
        bonds size if ``new_bond_dim`` is larger than product of the lower
        indices dimensions.

        Parameters
        ----------
        new_bond_dim : int
            The new maximum bond dimension to expand to.
        rand_strength : float, optional
            The strength of random noise to add to the new array entries,
            if any.
        inds_to_expand : sequence of str, optional
            The indices to expand, if not all.
        inplace : bool, optional
            Whether to expand this tensor network in place, or return a new
            one.

        Returns
        -------
        TensorNetworkGenIso
        """
        if inds_to_expand is not None:
            return super().expand_bond_dimension(
                new_bond_dim=new_bond_dim,
                rand_strength=rand_strength,
                inds_to_expand=inds_to_expand,
                inplace=inplace)

        tn = self if inplace else self.copy()

        tids_done = oset()
        inds_done = oset(tn.site_inds)
        tids_todo = tn._get_tids_from_inds(inds_done, 'any')

        # XXX: switch this logic to get_tree_span('CAP')? to
        # ensure topologically sorted order?
        while tids_todo:
            tid = tids_todo.popleft()
            t = tn.tensor_map[tid]

            if t.left_inds is not None:
                below_inds = oset(t.left_inds)
                above_inds = oset(t.inds) - below_inds
            else:
                below_inds, above_inds = oset(), oset()
                for ix in t.inds:
                    (below_inds if ix in inds_done else above_inds).add(ix)

            if len(above_inds) == 0:
                # top piece
                continue

            elif len(above_inds) == 1:
                # isometry, bond can expand
                ix, = above_inds
                cur_sz = t.ind_size(ix)
                rem_inds_sz = t.size // cur_sz

                # don't expand beyond product of lower index sizes
                new_sz = min(rem_inds_sz, new_bond_dim)
                if new_sz > cur_sz:
                    tn.expand_bond_dimension_(
                        new_bond_dim=new_sz,
                        rand_strength=rand_strength,
                        inds_to_expand=ix)

            elif len(above_inds) == len(below_inds):
                # unitary gate, maintain bond sizes
                for bix, aix in zip(below_inds, above_inds):
                    tn.expand_bond_dimension_(
                        new_bond_dim=t.ind_size(bix),
                        rand_strength=rand_strength,
                        inds_to_expand=aix
                    )

            else:
                raise NotImplementedError

            tids_done.add(tid)
            inds_done.update(above_inds)
            for tid_above in tn._get_tids_from_inds(above_inds, 'any'):
                if tid_above not in tids_done:
                    tids_todo.add(tid_above)

        return tn

    expand_bond_dimension_ = functools.partialmethod(
        expand_bond_dimension, inplace=True)


def calc_1d_unis_isos(sites, block_size, cyclic, group_from_right):
    """Given ``sites``, assumed to be in a 1D order, though not neccessarily
    contiguous, calculate unitary and isometry groupings::

               │         │ <- new grouped site
        ┐   ┌─────┐   ┌─────┐   ┌
        │   │ ISO │   │ ISO │   │
        ┘   └─────┘   └─────┘   └
        │   │..│..│   │..│..│   │
        ┌───┐  │  ┌───┐  │  ┌───┐
        │UNI│  │  │UNI│  │  │UNI│
        └───┘  │  └───┘  │  └───┘
        │   │ ... │   │ ... │   │
            ^^^^^^^ <- isometry groupings of size, block_size
        ^^^^^     ^^^^^ <- unitary groupings of size 2

    Parameters
    ----------
    sites : sequence of hashable
        The sites to apply a layer to.
    block_size : int
        How many sites to group together per isometry block. Note that
        currently the unitaries will only ever act on blocks of size 2 across
        isometry block boundaries.
    cyclic : bool
        Whether to apply disentangler / unitaries across the boundary. The
        isometries will never be applied across the boundary, but since they
        always form a tree such a bipartition is natural.
    group_from_right : bool
        Wether to group the sites starting from the left or right. This only
        matters if ``block_size`` does not divide the number of sites. Alternating
        between left and right more evenly tiles the unitaries and isometries,
        especially at lower layers.

    Returns
    -------
    unis : list[tuple]
        The unitary groupings.
    isos : list[tuple]
        The isometry groupings.
    """
    sites = tuple(sites)
    nsites = len(sites)

    # track this so we know neighboring sites
    ranks = {s: i for i, s in enumerate(sites)}

    # first we linearly partition the sites to form the isometry groups
    size = block_size * (nsites // block_size)
    if group_from_right:
        grouped = sites[-size:]
    else:
        grouped = sites[:size]
    isos = list(partition(block_size, grouped))

    # then we disentangle at the edges of the
    # isometries to form the unitaries
    unis = set()
    for iso in isos:
        # n.b. only when the groups are not adjacent (e.g. because the number
        # of sites doesn't divide) will there be a left (right) disentangler
        # which is not also a right (left) disentangler. In that case a site
        # can see 2 unitaries rather than the usual 1 unitary and 1 isometry:
        #            │
        #     ┐   ┌─────┐   ┌
        #     │   │ ISO │   │
        #     ┘   └─────┘   └
        #     │   │..│..│   │
        #     ┌───┐  │  ┌───┐
        #     │UNI│  │  │UNI│
        #     └───┘  │  └───┘
        #     │   │     │   │
        #     sl  si    sf  sr

        # attempt left disentangle
        si = iso[0]
        ri = ranks[si]
        if cyclic or ri > 0:
            sl = sites[ri - 1]
            unis.add((sl, si))

        # attempt right disentangle
        sf = iso[-1]
        rf = ranks[sf]
        if cyclic or rf < nsites - 1:
            sr = sites[(rf + 1) % nsites]
            unis.add((sf, sr))

    return sorted(unis), isos


class MERA(TensorNetwork1DVector, TensorNetworkGenIso):
    """Replacement class for ``MERA`` which uses the new infrastructure and
    thus has methods like ``compute_local_expectation``.
    """

    _EXTRA_PROPS = tuple(sorted(
        set(TensorNetwork1DVector._EXTRA_PROPS) |
        set(TensorNetworkGenIso._EXTRA_PROPS)
    ))
    _CONTRACT_STRUCTURED = False

    def __init__(self, *args, **kwargs):
        self._num_layers = None
        super().__init__(*args, **kwargs)

    @classmethod
    def from_fill_fn(
        cls,
        fill_fn,
        L,
        D,
        phys_dim=2,
        block_size=2,
        cyclic=True,
        uni_fill_fn=None,
        iso_fill_fn=None,
        cap_fill_fn=None,
        **kwargs
    ):
        """Create a 1D MERA using ``fill_fn(shape) -> array_like`` to fill the
        tensors.

        Parameters
        ----------
        fill_fn : callable
            A function which takes a shape and returns an array_like of that
            shape. You can override this specfically for the unitaries,
            isometries and cap tensors using the kwargs ``uni_fill_fn``,
            ``iso_fill_fn`` and ``cap_fill_fn``.
        L : int
            The number of sites.
        D : int
            The maximum bond dimension.
        phys_dim : int, optional
            The dimension of the physical indices.
        block_size : int, optional
            The size of the isometry blocks. Binary MERA is the default,
            ternary MERA is ``block_size=3``.
        cyclic : bool, optional
            Whether to apply disentangler / unitaries across the boundary. The
            isometries will never be applied across the boundary, but since
            they always form a tree such a bipartition is natural.
        uni_fill_fn : callable, optional
            A function which takes a shape and returns an array_like of that
            shape. This is used to fill the unitary tensors. If ``None`` then
            ``fill_fn`` is used.
        iso_fill_fn : callable, optional
            A function which takes a shape and returns an array_like of that
            shape. This is used to fill the isometry tensors. If ``None`` then
            ``fill_fn`` is used.
        cap_fill_fn : callable, optional
            A function which takes a shape and returns an array_like of that
            shape. This is used to fill the cap tensors. If ``None`` then
            ``fill_fn`` is used.
        kwargs
            Supplied to ``TensorNetworkGenIso.__init__``.
        """
        mera = cls.empty(sites=range(L), phys_dim=phys_dim, **kwargs)
        mera._L = L

        if uni_fill_fn is None:
            uni_fill_fn = fill_fn
        if iso_fill_fn is None:
            iso_fill_fn = fill_fn
        if cap_fill_fn is None:
            cap_fill_fn = iso_fill_fn

        for lyr in itertools.count():
            remaining_sites = sorted(mera._open_upper_sites)

            if len(remaining_sites) <= block_size + 1:
                # can terminate with a 'cap'
                mera.layer_gate_fill_fn(
                    cap_fill_fn, "cap", remaining_sites, D, tags=f'LAYER{lyr}',
                )
                break

            # else add a disentangling and grouping layer
            uni_groups, iso_groups = calc_1d_unis_isos(
                remaining_sites, block_size, cyclic, group_from_right=lyr % 2,
            )
            for uni_sites in uni_groups:
                mera.layer_gate_fill_fn(
                    uni_fill_fn, "uni", uni_sites, D, tags=f'LAYER{lyr}',
                )
            for iso_sites in iso_groups:
                mera.layer_gate_fill_fn(
                    iso_fill_fn, "iso", iso_sites, D, tags=f'LAYER{lyr}',
                )

        mera._num_layers = lyr + 1

        return mera

    @classmethod
    def rand(
        cls,
        L,
        D,
        seed=None,
        block_size=2,
        phys_dim=2,
        cyclic=True,
        isometrize_method="qr",
        **kwargs
    ):
        """Return a random (optionally isometrized) MERA.

        Parameters
        ----------
        L : int
            The number of sites.
        D : int
            The maximum bond dimension.
        seed : int, optional
            A random seed.
        block_size : int, optional
            The size of the isometry blocks. Binary MERA is the default,
            ternary MERA is ``block_size=3``.
        phys_dim : int, optional
            The dimension of the physical indices.
        cyclic : bool, optional
            Whether to apply disentangler / unitaries across the boundary. The
            isometries will never be applied across the boundary, but since
            they always form a tree such a bipartition is natural.
        isometrize_method : str or None, optional
            If given, the method to use to isometrize the MERA. If ``None``
            then the MERA is not isometrized.
        """
        import numpy as np
        rng = np.random.default_rng(seed)
        mera = cls.from_fill_fn(
            lambda shape: rng.normal(size=shape),
            L, D, block_size, phys_dim, cyclic, **kwargs,
        )
        if isometrize_method is not None:
            mera.isometrize_(isometrize_method)
        return mera

    @property
    def num_layers(self):
        return self._num_layers
